import numpy as np
from astropy.io import fits
import pytest

from stcal.jump.twopoint_difference import calc_med_first_diffs, find_crs

DQFLAGS = {"JUMP_DET": 4, "SATURATED": 2, "DO_NOT_USE": 1}


@pytest.fixture()
def setup_cube():

    def _cube(ngroups, nints=1, nrows=204, ncols=204, readnoise=10):
        rej_threshold = 3
        nframes = 1
        data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
        read_noise = np.full((nrows, ncols), readnoise, dtype=np.float32)
        gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)

        return data, gdq, nframes, read_noise, rej_threshold

    return _cube


def test_varying_groups(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, nints=1, nrows=2, ncols=2, readnoise=8)
    data[0, :, 0, 0] = [10, 20, 30, 530, 540]
    data[0, :, 0, 1] = [10, 20, 30, 530, np.nan]
    data[0, :, 1, 0] = [10, 20, 530, np.nan, np.nan]
    data[0, :, 1, 1] = [10, 520, np.nan, np.nan, np.nan]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.array_equal(out_gdq[0, :, 0, 0], [0, 0, 0, 4, 0])
    assert np.array_equal(out_gdq[0, :, 0, 1], [0, 0, 0, 4, 0])
    assert np.array_equal(out_gdq[0, :, 1, 0], [0, 0, 4, 0, 0])
    assert np.array_equal(out_gdq[0, :, 1, 1], [0, 0, 0, 0, 0])


def test_multint_pixel(setup_cube):
    ngroups=4
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, nints=7, nrows=2, ncols=2, readnoise=8)
    data[0, :, 0, 0] = (-24,   -15,     0,    13)
    data[1, :, 0, 0] = (-24,   -11,     6,    21)
    data[2, :, 0, 0] = (-40,   -28,   -24,    -4)
    data[3, :, 0, 0] = (-11,     3,    11,    24)
    data[4, :, 0, 0] = (-43 ,  -24,   -12,     1)
    data[5, :, 0, 0] = (-45,  8537, 17380, 17437)
    data[6, :, 0, 0] = (-178,  -156,  -139,  -125)
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert (np.array_equal([0, 4, 4, 4], out_gdq[5, :, 0, 0]))



def test_nocrs_noflux(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )

    assert np.max(out_gdq) == 0  # no CR found


def test_5grps_cr3_noflux(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)

    data[0, 0:2, 100, 100] = 10.0
    data[0, 2:5, 100, 100] = 1000
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.argmax(out_gdq[0, :, 100, 100]) == 2  # find the CR in the expected group
    data[0, 0, 100, 100] = 10.0
    data[0, 1:6, 100, 100] = 1000

def test_4grps_2ints_cr2_noflux(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, nints=2, ncols=2, nrows=2)
    data[0, 1, 1, 1] = 5
    data[1, 0, 1, 1] = 10.0
    data[1, 1:6, 1, 1] = 1000
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(data, gdq, read_noise, rej_threshold,
                                                     rej_threshold, rej_threshold, nframes,
                                                     False, 200, 10, DQFLAGS)
    assert(4 == np.max(out_gdq))  # a CR was found
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.argmax(out_gdq[1, :, 1, 1]) == 1  # find the CR in the expected group
    assert(1 == np.argmax(out_gdq[1, :, 1, 1]))  # find the CR in the expected group

def test_6grps_negative_differences_zeromedian(setup_cube):
    ngroups = 6
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)

    data[0, 0, 100, 100] = 100
    data[0, 1, 100, 100] = 90
    data[0, 2, 100, 100] = 95
    data[0, 3, 100, 100] = 105
    data[0, 4, 100, 100] = 100
    data[0, 5, 100, 100] = 100
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 0  # no CR was found


def test_5grps_cr2_negjumpflux(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)

    data[0, 0, 100, 100] = 1000.0
    data[0, 1:6, 100, 100] = 10
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.argmax(out_gdq[0, :, 100, 100]) == 1  # find the CR in the expected group


def test_3grps_cr2_noflux(setup_cube):
    ngroups = 3
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    data[0, 0, 100, 100] = 10.0
    data[0, 1:4, 100, 100] = 1000
    data[0, 0, 99, 99] = 10.0
    data[0, 2:4, 99, 99] = 1000
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    #    assert(1,np.argmax(out_gdq[0, :, 100, 100]))  # find the CR in the expected group
    assert(np.array_equal([0, 4, 0], out_gdq[0, :, 100, 100]))
    assert (np.array_equal([0, 0, 4], out_gdq[0, :, 99, 99]))

def test_2ints_2grps_noflux(setup_cube):
    ngroups = 2
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, nints=2, ncols=2, nrows=2)
    data[0, 0, 1, 1] = 10.0
    data[0, 1:3, 1, 1] = 1000
    data[1, 0, 0, 0] = 10.0
    data[1, 1:3, 0, 0] = 1000
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(data, gdq, read_noise, rej_threshold,
                                                     rej_threshold, rej_threshold, nframes,
                                                     False, 200, 10, DQFLAGS,
                                                                         minimum_groups=2)
 #   assert(4 == np.max(out_gdq))  # a CR was found
    #    assert(1,np.argmax(out_gdq[0, :, 100, 100]))  # find the CR in the expected group
    assert(np.array_equal([0, 4], out_gdq[0, :, 1, 1]))
    assert (np.array_equal([0, 4], out_gdq[1, :, 0, 0]))

def test_4grps_cr2_noflux(setup_cube):
    ngroups = 4
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    data[0, 0, 100, 100] = 10.0
    data[0, 1:4, 100, 100] = 1000
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.argmax(out_gdq[0, :, 100, 100]) == 1  # find the CR in the expected group


def test_6grps_cr2_nframe2(setup_cube):
    ngroups = 6
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, nrows=2, ncols=2)
    nframes = 2
    data[0, 0, 1, 1] = 10.0
    data[0, 1, 1, 1] = 500
    data[0, 2, 1, 1] = 1002
    data[0, 3, 1, 1] = 1001
    data[0, 4, 1, 1] = 1005
    data[0, 5, 1, 1] = 1015
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert (np.array_equal([0, 4, 4, 0, 0, 0], out_gdq[0, :, 1, 1]))
    assert (np.max(out_gdq[0, :, 0, 0]) == 0)
    assert (np.max(out_gdq[0, :, 1, 0]) == 0)
    assert (np.max(out_gdq[0, :, 0, 1]) == 0)


def test_4grps_twocrs_2nd_4th(setup_cube):
    ngroups = 4
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    nframes = 1
    data[0, 0, 100, 100] = 10.0
    data[0, 1, 100, 100] = 60
    data[0, 2, 100, 100] = 60
    data[0, 3, 100, 100] = 115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found


def test_5grps_twocrs_2nd_5th(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    nframes = 1
    data[0, 0, 100, 100] = 10.0
    data[0, 1, 100, 100] = 60
    data[0, 2, 100, 100] = 60
    data[0, 3, 100, 100] = 60
    data[0, 4, 100, 100] = 115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 4], out_gdq[0, :, 100, 100])


def test_5grps_twocrs_2nd_5thbig(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    nframes = 1
    data[0, 0, 100, 100] = 10.0
    data[0, 1, 100, 100] = 600
    data[0, 2, 100, 100] = 600
    data[0, 3, 100, 100] = 600
    data[0, 4, 100, 100] = 2115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 4], out_gdq[0, :, 100, 100])


def test_10grps_twocrs_2nd_8th_big(setup_cube):
    ngroups = 10
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    nframes = 1
    data[0, 0, 100, 100] = 10.0
    data[0, 1, 100, 100] = 60
    data[0, 2, 100, 100] = 60
    data[0, 3, 100, 100] = 60
    data[0, 4, 100, 100] = 60
    data[0, 5, 100, 100] = 60
    data[0, 6, 100, 100] = 60
    data[0, 7, 100, 100] = 2115
    data[0, 8, 100, 100] = 2115
    data[0, 9, 100, 100] = 2115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 0, 0, 0, 4, 0, 0], out_gdq[0, :, 100, 100])


def test_10grps_twocrs_10percenthit(setup_cube):
    ngroups = 10
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    nframes = 2
    data[0:200, 0, 100, 100] = 10.0
    data[0:200, 1, 100, 100] = 60
    data[0:200, 2, 100, 100] = 60
    data[0:200, 3, 100, 100] = 60
    data[0:200, 4, 100, 100] = 60
    data[0:200, 5, 100, 100] = 60
    data[0:200, 6, 100, 100] = 60
    data[0:200, 7, 100, 100] = 2115
    data[0:200, 8, 100, 100] = 2115
    data[0:200, 9, 100, 100] = 2115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 0, 0, 0, 4, 0, 0], out_gdq[0, :, 100, 100])


def test_5grps_twocrs_2nd_5thbig_nframes2(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10 * np.sqrt(2))
    nframes = 2
    data[0, 0, 100, 100] = 10.0
    data[0, 1, 100, 100] = 60
    data[0, 2, 100, 100] = 60
    data[0, 3, 100, 100] = 60
    data[0, 4, 100, 100] = 2115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 4], out_gdq[0, :, 100, 100])


def test_6grps_twocrs_2nd_5th(setup_cube):
    ngroups = 6
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    nframes = 1
    data[0, 0, 100, 100] = 10.0
    data[0, 1, 100, 100] = 60
    data[0, 2, 100, 100] = 60
    data[0, 3, 100, 100] = 60
    data[0, 4, 100, 100] = 115
    data[0, 5, 100, 100] = 115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 4, 0], out_gdq[0, :, 100, 100])


def test_6grps_twocrs_2nd_5th_nframes2(setup_cube):
    ngroups = 6
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10 * np.sqrt(2))
    nframes = 2
    data[0, 0, 100, 100] = 10.0
    data[0, 1, 100, 100] = 60
    data[0, 2, 100, 100] = 60
    data[0, 3, 100, 100] = 60
    data[0, 4, 100, 100] = 115
    data[0, 5, 100, 100] = 115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 4, 0], out_gdq[0, :, 100, 100])


def test_6grps_twocrs_twopixels_nframes2(setup_cube):
    ngroups = 6
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10 * np.sqrt(2))
    nframes = 2
    data[0, 0, 100, 100] = 10.0
    data[0, 1, 100, 100] = 60
    data[0, 2, 100, 100] = 60
    data[0, 3, 100, 100] = 60
    data[0, 4, 100, 100] = 115
    data[0, 5, 100, 100] = 115
    data[0, 0, 200, 100] = 10.0
    data[0, 1, 200, 100] = 10.0
    data[0, 2, 200, 100] = 60
    data[0, 3, 200, 100] = 60
    data[0, 4, 200, 100] = 115
    data[0, 5, 200, 100] = 115
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 4, 0], out_gdq[0, :, 100, 100])
    assert np.array_equal([0, 0, 4, 0, 4, 0], out_gdq[0, :, 200, 100])


def test_5grps_cr2_negslope(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups)
    nframes = 1
    data[0, 0, 1, 1] = 100.0
    data[0, 1, 1, 1] = 0
    data[0, 2, 1, 1] = -200
    data[0, 3, 1, 1] = -260
    data[0, 4, 1, 1] = -360
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 0, 4, 0, 0], out_gdq[0, :, 1, 1])


def test_6grps_1cr(setup_cube):
    ngroups = 6
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    data[0, 5, 100, 100] = 1146
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert out_gdq[0, 5, 100, 100] == 4


def test_7grps_1cr(setup_cube):
    ngroups = 7
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    data[0, 5, 100, 100] = 60
    data[0, 6, 100, 100] = 1160
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert out_gdq[0, 6, 100, 100] == 4


def test_8grps_1cr(setup_cube):
    ngroups = 8
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    data[0, 5, 100, 100] = 60
    data[0, 6, 100, 100] = 1160
    data[0, 7, 100, 100] = 1175
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert out_gdq[0, 6, 100, 100] == 4


def test_9grps_1cr_1sat(setup_cube):
    ngroups = 9
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10, nrows=2, ncols=2)
    nframes = 1
    data[0, 0, 1, 1] = 0
    data[0, 1, 1, 1] = 10
    data[0, 2, 1, 1] = 21
    data[0, 3, 1, 1] = 33
    data[0, 4, 1, 1] = 46
    data[0, 5, 1, 1] = 60
    data[0, 6, 1, 1] = 1160
    data[0, 7, 1, 1] = 1175
    data[0, 8, 1, 1] = 6175
    gdq[0, 8, 1, 1] = DQFLAGS["SATURATED"]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert out_gdq[0, 6, 1, 1] == 4


def test_10grps_1cr_2sat(setup_cube):
    ngroups = 10
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    data[0, 5, 100, 100] = 60
    data[0, 6, 100, 100] = 1160
    data[0, 7, 100, 100] = 1175
    data[0, 8, 100, 100] = 6175
    data[0, 9, 100, 100] = 6175
    gdq[0, 8, 100, 100] = DQFLAGS["SATURATED"]
    gdq[0, 9, 100, 100] = DQFLAGS["SATURATED"]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert out_gdq[0, 6, 100, 100] == 4


def test_11grps_1cr_3sat(setup_cube):
    ngroups = 11
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10, nrows=2, ncols=2)
    nframes = 1
    data[0, 0, 1, 1] = 0
    data[0, 1, 1, 1] = 20
    data[0, 2, 1, 1] = 39
    data[0, 3, 1, 1] = 57
    data[0, 4, 1, 1] = 74
    data[0, 5, 1, 1] = 90
    data[0, 6, 1, 1] = 1160
    data[0, 7, 1, 1] = 1175
    data[0, 8, 1, 1] = 6175
    data[0, 9, 1, 1] = 6175
    data[0, 10, 1, 1] = 6175
    gdq[0, 8, 1, 1] = DQFLAGS["SATURATED"]
    gdq[0, 9, 1, 1] = DQFLAGS["SATURATED"]
    gdq[0, 10, 1, 1] = DQFLAGS["SATURATED"]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert out_gdq[0, 6, 1, 1] == 4


def test_11grps_0cr_3donotuse(setup_cube):
    ngroups = 11
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 18
    data[0, 2, 100, 100] = 39
    data[0, 3, 100, 100] = 57
    data[0, 4, 100, 100] = 74
    data[0, 5, 100, 100] = 90
    data[0, 6, 100, 100] = 115
    data[0, 7, 100, 100] = 131
    data[0, 8, 100, 100] = 150
    data[0, 9, 100, 100] = 6175
    data[0, 10, 100, 100] = 6175
    gdq[0, 0, 100, 100] = DQFLAGS["DO_NOT_USE"]
    gdq[0, 9, 100, 100] = DQFLAGS["DO_NOT_USE"]
    gdq[0, 10, 100, 100] = DQFLAGS["DO_NOT_USE"]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.array_equal([0, 0, 0, 0, 0, 0, 0, 0], out_gdq[0, 1:-2, 100, 100])


def test_5grps_nocr(setup_cube):
    ngroups = 6
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )


def test_6grps_nocr(setup_cube):
    ngroups = 6
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    data[0, 5, 100, 100] = 60
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )


def test_10grps_cr2_gt3sigma(setup_cube):
    ngroups = 10
    crmag = 16
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1:11, 100, 100] = crmag
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 0, 0, 0, 0, 0, 0], out_gdq[0, :, 100, 100])


def test_10grps_cr2_3sigma_nocr(setup_cube):
    ngroups = 10
    crmag = 15
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1:11, 100, 100] = crmag
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 0  # a CR was found
    assert np.array_equal([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], out_gdq[0, :, 100, 100])


def test_10grps_cr2_gt3sigma_2frames(setup_cube):
    ngroups = 10
    crmag = 16
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5 * np.sqrt(2))
    nframes = 2
    data[0, 0, 100, 100] = 0
    data[0, 1:11, 100, 100] = crmag
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 0, 0, 0, 0, 0, 0], out_gdq[0, :, 100, 100])


def test_10grps_cr2_gt3sigma_2frames_offdiag(setup_cube):
    ngroups = 10
    crmag = 16
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5 * np.sqrt(2))
    nframes = 2
    data[0, 0, 100, 110] = 0
    data[0, 1:11, 100, 110] = crmag
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 4  # a CR was found
    assert np.array_equal([0, 4, 0, 0, 0, 0, 0, 0, 0, 0], out_gdq[0, :, 100, 110])


def test_10grps_cr2_3sigma_2frames_nocr(setup_cube):
    ngroups = 10
    crmag = 15
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5 * np.sqrt(2))
    nframes = 2
    data[0, 0, 100, 100] = 0
    data[0, 1:11, 100, 100] = crmag
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 0  # a CR was found
    assert np.array_equal([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], out_gdq[0, :, 100, 100])


def test_10grps_nocr_2pixels_sigma0(setup_cube):
    ngroups = 10
    crmag = 15
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5 * np.sqrt(2))
    nframes = 1
    data[0, 0, 100, 100] = crmag
    data[0, 1:11, 100, 100] = crmag
    read_noise[50, 50] = 0.0
    read_noise[60, 60] = 0.0
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.max(out_gdq) == 0  # no CR was found


def test_5grps_satat4_crat3(setup_cube):
    ngroups = 5
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, nrows=2, ncols=2, readnoise=5 * np.sqrt(2))
    nframes = 1
    data[0, 0, 1, 1] = 10000
    data[0, 1, 1, 1] = 20000
    data[0, 2, 1, 1] = 60000
    data[0, 3, 1, 1] = 61000
    data[0, 4, 1, 1] = 61000
    gdq[0, 3, 1, 1] = DQFLAGS['SATURATED']
    gdq[0, 4, 1, 1] = DQFLAGS['SATURATED']

    data[0, 0, 0, 1] = 59800
    data[0, 1, 0, 1] = 59900
    data[0, 2, 0, 1] = 60000
    data[0, 3, 0, 1] = 61000
    data[0, 4, 0, 1] = 61000
    gdq[0, 3, 0, 1] = DQFLAGS['SATURATED']
    gdq[0, 4, 0, 1] = DQFLAGS['SATURATED']
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(data, gdq, read_noise, rej_threshold,
                                                     rej_threshold, rej_threshold, nframes,
                                                     False, 200, 10, DQFLAGS)
    # assert(4 == np.max(out_gdq))  # no CR was found
    result = out_gdq[0, :, 1, 1]
    assert np.array_equal(
        [0, 0, DQFLAGS['JUMP_DET'], DQFLAGS['SATURATED'], DQFLAGS['SATURATED']],
        out_gdq[0, :, 1, 1])
    assert np.array_equal(
        [0, 0, 0, DQFLAGS['SATURATED'], DQFLAGS['SATURATED']],
        out_gdq[0, :, 0, 1])

def test_6grps_satat6_crat1(setup_cube):
    ngroups = 6
    # crmag = 1000
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5 * np.sqrt(2))
    nframes = 1
    data[0, 0, 100, 100] = 10000
    data[0, 1, 100, 100] = 35000  # CR
    data[0, 2, 100, 100] = 40005
    data[0, 3, 100, 100] = 45029
    data[0, 4, 100, 100] = 50014
    data[0, 5, 100, 101] = 61000
    data[0, 0, 100, 101] = 10000
    data[0, 1, 100, 101] = 15001
    data[0, 2, 100, 101] = 20003
    data[0, 3, 100, 101] = 25006
    data[0, 4, 100, 101] = 30010
    data[0, 5, 100, 101] = 35015
    gdq[0, 5, 100, 100] = DQFLAGS["SATURATED"]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    # assert(4 == np.max(out_gdq))  # no CR was found
    assert np.array_equal([0, DQFLAGS["JUMP_DET"], 0, 0, 0, DQFLAGS["SATURATED"]], out_gdq[0, :, 100, 100])


@pytest.mark.xfail()
def test_6grps_satat6_crat1_flagadjpixels(setup_cube):
    ngroups = 6
    # crmag = 1000
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5 * np.sqrt(2))
    nframes = 1
    data[0, 0, 100, 100] = 10000
    data[0, 1, 100, 100] = 35000  # CR
    data[0, 2, 100, 100] = 40005
    data[0, 3, 100, 100] = 45029
    data[0, 4, 100, 100] = 50014
    data[0, 5, 100, 101] = 61000
    data[0, 0, 100, 101] = 10000
    data[0, 1, 100, 101] = 15001
    data[0, 2, 100, 101] = 20003
    data[0, 3, 100, 101] = 25006
    data[0, 4, 100, 101] = 30010
    data[0, 5, 100, 101] = 35015
    gdq[0, 5, 100, 100] = DQFLAGS["SATURATED"]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    # assert(4 == np.max(out_gdq))  # no CR was found
    assert np.array_equal([0, DQFLAGS["JUMP_DET"], 0, 0, 0, DQFLAGS["SATURATED"]], out_gdq[0, :, 100, 100])
    assert np.array_equal([0, DQFLAGS["JUMP_DET"], 0, 0, 0, DQFLAGS["SATURATED"]], out_gdq[0, :, 99, 100])


def test_10grps_satat8_crsat3and6(setup_cube):
    ngroups = 10
    # crmag = 1000
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5 * np.sqrt(2))
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 5000
    data[0, 2, 100, 100] = 15000  # CR
    data[0, 3, 100, 100] = 20000
    data[0, 4, 100, 100] = 25000
    data[0, 5, 100, 100] = 40000  # CR
    data[0, 6, 100, 100] = 45000
    data[0, 7:11, 100, 100] = 61000
    gdq[0, 7:11, 100, 100] = DQFLAGS["SATURATED"]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    # assert(4 == np.max(out_gdq))  # no CR was found
    assert np.array_equal(
        [
            0,
            0,
            DQFLAGS["JUMP_DET"],
            0,
            0,
            DQFLAGS["JUMP_DET"],
            0,
            DQFLAGS["SATURATED"],
            DQFLAGS["SATURATED"],
            DQFLAGS["SATURATED"],
        ],
        out_gdq[0, :, 100, 100],
    )


def test_median_with_saturation(setup_cube):
    ngroups = 10
    # crmag = 1000
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, nrows=2, ncols=2, readnoise=5 * np.sqrt(2))
    nframes = 1
    data[0, 0, 1, 1] = 0
    data[0, 1, 1, 1] = 4700
    data[0, 2, 1, 1] = 9400
    data[0, 3, 1, 1] = 14300
    data[0, 4, 1, 1] = 19100
    data[0, 5, 1, 1] = 40000  # CR
    data[0, 6, 1, 1] = 44850
    data[0, 7, 1, 1] = 49900
    data[0, 8:10, 1, 1] = 60000
    gdq[0, 7:10, 1, 1] = DQFLAGS['SATURATED']
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(data, gdq, read_noise, rej_threshold,
                                                     rej_threshold, rej_threshold, nframes,
                                                     False, 200, 10, DQFLAGS)
    gdq_value = out_gdq[0, :, 1, 1]
    assert (np.array_equal([0, 0, 0, 0, 0, 4, 0, 2, 2, 2], out_gdq[0, :, 1, 1]))


def test_median_with_saturation_odd_number_final_difference(setup_cube):
    ngroups = 9
    # crmag = 1000
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=5 * np.sqrt(2))
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 4500
    data[0, 2, 100, 100] = 9100
    data[0, 3, 100, 100] = 13800
    data[0, 4, 100, 100] = 18600
    data[0, 5, 100, 100] = 40000  # CR
    data[0, 6, 100, 100] = 44850
    data[0, 7, 100, 100] = 49900
    data[0, 8:9, 100, 100] = 60000
    gdq[0, 6:9, 100, 100] = DQFLAGS["SATURATED"]
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert np.array_equal([0, 0, 0, 0, 0, 4, 2, 2, 2], out_gdq[0, :, 100, 100])


def test_first_last_group(setup_cube):
    ngroups = 7
    nframes = 1
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=25.0)

    #  set up the data so that if the first and last group are used in jump
    #  detection it would cause a jump to be detected between group 1-2
    #  and group 6-7. Add a jump between 3 and 4 just to make sure jump detection is working
    #  set group 1 to be 10,000
    data[0, 0, 100, 100] = 10000.0
    #  set groups 1,2 - to be around 30,000
    data[0, 1, 100, 100] = 30000.0
    data[0, 2, 100, 100] = 30020.0
    #  set up a jump to make sure it is detected
    data[0, 3, 100, 100] = 40000.0
    data[0, 4, 100, 100] = 40020.0
    data[0, 5, 100, 100] = 40040.0
    #  set group 6 to be 50,000
    data[0, 6, 100, 100] = 50000.0

    gdq[0, 0, 100, 100] = DQFLAGS["DO_NOT_USE"]
    gdq[0, 6, 100, 100] = DQFLAGS["DO_NOT_USE"]
    outgdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )

    assert outgdq[0, 0, 100, 100] == DQFLAGS["DO_NOT_USE"]
    assert outgdq[0, 6, 100, 100] == DQFLAGS["DO_NOT_USE"]
    assert outgdq[0, 3, 100, 100] == DQFLAGS["JUMP_DET"]


def test_2group(setup_cube):
    # test should not find a CR, can't do it with only one difference.
    ngroups = 2
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=25.0)

    data[0, 0, 0, 0] = 10000.0
    #  set groups 1,2 - to be around 30,000
    data[0, 1, 0, 0] = 30000.0

    outgdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert outgdq[0, 1, 0, 0] == 0
    assert outgdq[0, 0, 0, 0] == 0


def test_4group(setup_cube):
    ngroups = 4
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=25.0)

    data[0, 0, 0, 0] = 10000.0
    #  set groups 1,2 - to be around 30,000
    data[0, 1, 0, 0] = 30000.0
    data[0, 2, 0, 0] = 30020.0
    data[0, 3, 0, 0] = 30000.0

    outgdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )
    assert outgdq[0, 1, 0, 0] == 4


def test_first_last_4group(setup_cube):
    ngroups = 4
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=25.0)

    #  set up the data so that if the first and last group are used in jump
    #  detection it would cause a jump to be detected between group 0-1.
    data[0, 0, 0, 0] = 10000.0
    #  set groups 1,2 - to be around 30,000
    data[0, 1, 0, 0] = 30000.0
    data[0, 2, 0, 0] = 30020.0
    data[0, 3, 0, 0] = 30000.0
    # treat as MIRI data with first and last flagged
    gdq[0, 0, :, :] = DQFLAGS["DO_NOT_USE"]
    gdq[0, 3, :, :] = DQFLAGS["DO_NOT_USE"]
    outgdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )

    assert outgdq[0, 0, 0, 0] == DQFLAGS["DO_NOT_USE"]
    assert outgdq[0, 3, 0, 0] == DQFLAGS["DO_NOT_USE"]
    assert outgdq[0, 1, 0, 0] == 0


def test_first_last_3group(setup_cube):
    ngroups = 3
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=25.0)

    #  set up the data so that if the first and last group are used in jump
    #  detection it would cause a jump to be detected between group 1-2
    #  and group 6-7. Add a jump between 3 and 4 just to make sure jump detection is working
    #  set group 1 to be 10,000
    data[0, 0, 0, 0] = 10000.0
    data[0, 1, 0, 0] = 10100.0
    data[0, 2, 0, 0] = 30020.0

    gdq[0, 2, 0, 0] = DQFLAGS["DO_NOT_USE"]  # only flag the last group
    outgdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data, gdq, read_noise, rej_threshold, rej_threshold, rej_threshold, nframes, False, 200, 10, DQFLAGS
    )

    assert outgdq[0, 0, 0, 0] == 0
    assert outgdq[0, 2, 0, 0] == DQFLAGS["DO_NOT_USE"]
    assert outgdq[0, 1, 0, 0] == 0


def test_10grps_1cr_afterjump(setup_cube):
    ngroups = 10
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    data[0, 5, 100, 100] = 60
    data[0, 6, 100, 100] = 1160
    data[0, 7, 100, 100] = 1175
    data[0, 8, 100, 100] = 1190
    data[0, 9, 100, 100] = 1209

    after_jump_flag_e1 = 0.0
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data,
        gdq,
        read_noise,
        rej_threshold,
        rej_threshold,
        rej_threshold,
        nframes,
        False,
        200,
        10,
        DQFLAGS,
        after_jump_flag_e1=after_jump_flag_e1,
        after_jump_flag_n1=10,
    )
    # all groups after CR should be flagged
    for k in range(6, 10):
        assert out_gdq[0, k, 100, 100] == 4, f"after jump flagging failed in group {k}"


def test_10grps_1cr_afterjump_2group(setup_cube):
    ngroups = 10
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    data[0, 5, 100, 100] = 60
    data[0, 6, 100, 100] = 1160
    data[0, 7, 100, 100] = 1175
    data[0, 8, 100, 100] = 1190
    data[0, 9, 100, 100] = 1209

    after_jump_flag_e1 = 0.0
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data,
        gdq,
        read_noise,
        rej_threshold,
        rej_threshold,
        rej_threshold,
        nframes,
        False,
        200,
        10,
        DQFLAGS,
        after_jump_flag_e1=after_jump_flag_e1,
        after_jump_flag_n1=2,
    )

    # 2 groups after CR should be flagged
    for k in range(6, 9):
        assert out_gdq[0, k, 100, 100] == 4, f"after jump flagging failed in group {k}"

    # rest not flagged
    for k in range(9, 10):
        assert out_gdq[0, k, 100, 100] == 0, f"after jump flagging incorrect in group {k}"


def test_10grps_1cr_afterjump_toosmall(setup_cube):
    ngroups = 10
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 21
    data[0, 3, 100, 100] = 33
    data[0, 4, 100, 100] = 46
    data[0, 5, 100, 100] = 60
    data[0, 6, 100, 100] = 1160
    data[0, 7, 100, 100] = 1175
    data[0, 8, 100, 100] = 1190
    data[0, 9, 100, 100] = 1209

    after_jump_flag_e1 = 10000.0
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data,
        gdq,
        read_noise,
        rej_threshold,
        rej_threshold,
        rej_threshold,
        nframes,
        False,
        200,
        10,
        DQFLAGS,
        after_jump_flag_e1=after_jump_flag_e1,
        after_jump_flag_n1=10,
    )
    # all groups after CR should be flagged
    for k in range(7, 10):
        assert out_gdq[0, k, 100, 100] == 0, f"after jump flagging incorrect in group {k}"


def test_10grps_1cr_afterjump_twothresholds(setup_cube):
    ngroups = 10
    data, gdq, nframes, read_noise, rej_threshold = setup_cube(ngroups, readnoise=10)
    nframes = 1
    data[0, 0, 100, 100] = 0
    data[0, 1, 100, 100] = 10
    data[0, 2, 100, 100] = 121
    data[0, 3, 100, 100] = 133
    data[0, 4, 100, 100] = 146
    data[0, 5, 100, 100] = 160
    data[0, 6, 100, 100] = 1160
    data[0, 7, 100, 100] = 1175
    data[0, 8, 100, 100] = 1190
    data[0, 9, 100, 100] = 1209

    after_jump_flag_e1 = 500.0
    after_jump_flag_e2 = 10.0
    out_gdq, row_below_gdq, rows_above_gdq, total_crs, stddev = find_crs(
        data,
        gdq,
        read_noise,
        rej_threshold,
        rej_threshold,
        rej_threshold,
        nframes,
        False,
        200,
        10,
        DQFLAGS,
        after_jump_flag_e1=after_jump_flag_e1,
        after_jump_flag_n1=10,
        after_jump_flag_e2=after_jump_flag_e2,
        after_jump_flag_n2=2,
    )
    # 2 groups after CR should be flagged
    for k in range(2, 5):
        assert out_gdq[0, k, 100, 100] == 4, f"after jump flagging incorrect in group {k}"

    # all groups after CR should be flagged
    for k in range(6, 10):
        assert out_gdq[0, k, 100, 100] == 4, f"after jump flagging incorrect in group {k}"


def test_median_func():
    """
    Test the function `calc_med_first_diffs` that computes median of pixels.
    Ensure that the correct treatment based on number of non-nan diffs
    is being done, and that it works for individual pixels as well as
    pixels embedded in 3d arrays, and that it works for arrays with or
    without nans (which represent masked pixels)."""

    # single pix with 5 good diffs, should clip 1 pix and return median
    # 1d, no nans
    arr = np.array([1.0, 2.0, 3.0, 4.0, 5])
    assert calc_med_first_diffs(arr) == 2.5
    # 3d array, no nans
    arr = np.zeros(5 * 2 * 2).reshape(1, 5, 2, 2)
    arr[0, :, 0, 0] = np.array([1.0, 2.0, 3.0, 5.0, 4])
    arr[0, 1, 0, 1] = 1.0
    result = calc_med_first_diffs(arr)
    assert result[0, 0] == 2.5
    # 1d, with nans
    arr = np.array([1.0, 2.0, 3.0, np.nan, 4.0, 5, np.nan])
    assert calc_med_first_diffs(arr) == 2.5
    # 3d, with nans
    arr = np.zeros(7 * 2 * 2).reshape(1, 7, 2, 2)
    arr[0, :, 0, 0] = np.array([1.0, 2.0, 3.0, np.nan, 4.0, 5, np.nan])
    assert calc_med_first_diffs(arr)[0, 0] == 2.5

    # single pix with exactly 4 good diffs, should also clip 1 pix and return median
    # 1d, no nans
    arr = np.array([1.0, 2.0, 3.0, 4.0])
    assert calc_med_first_diffs(arr) == 2
    # 3d array, no nans
    arr = np.zeros(4 * 2 * 2).reshape(1, 4, 2, 2)
    arr[0, :, 0, 0] = np.array([1.0, 2.0, 3.0, 4.0])
    assert calc_med_first_diffs(arr)[0, 0] == 2
    # 1d, with nans
    arr = np.array([1.0, 2.0, 3.0, np.nan, 4.0, np.nan])
    assert calc_med_first_diffs(arr) == 2
    # 3d, with nans
    arr = np.zeros(6 * 2 * 2).reshape(1, 6, 2, 2)
    arr[0, :, 0, 0] = np.array([1.0, 2.0, 3.0, np.nan, 4.0, np.nan])
    assert calc_med_first_diffs(arr)[0, 0] == 2

    # single pix with exactly 3 good diffs, should compute median without clipping
    arr = np.array([1.0, 2.0, 3.0])
    assert calc_med_first_diffs(arr) == 2
    # 3d array, no nans
    arr = np.zeros(3 * 2 * 2).reshape(1, 3, 2, 2)
    arr[0, :, 0, 0] = np.array([1.0, 2.0, 3.0])
    assert calc_med_first_diffs(arr)[0, 0] == 2
    # 1d, with nans
    arr = np.array([1.0, 2.0, 3.0, np.nan, np.nan])
    assert calc_med_first_diffs(arr) == 2
    # 3d, with nans
    arr = np.zeros(5 * 2 * 2).reshape(1, 5, 2, 2)
    arr[0, :, 0, 0] = np.array([1.0, 2.0, 3.0, np.nan, np.nan])
    assert calc_med_first_diffs(arr)[0, 0] == 2

    # # single pix with exactly 2 good diffs, should return the element with the minimum abs val
    arr = np.array([-1.0, -2.0])
    assert calc_med_first_diffs(arr) == -1
    # 3d array, no nans
    arr = np.zeros(2 * 2 * 2).reshape(1, 2, 2, 2)
    arr[0, :, 0, 0] = np.array([-1.0, -2.0])
    assert calc_med_first_diffs(arr)[0, 0] == -1
    # 1d, with nans
    arr = np.array([-1.0, -2.0, np.nan, np.nan])
    assert calc_med_first_diffs(arr) == -1
    # 3d, with nans
    arr = np.zeros(4 * 2 * 2).reshape(1, 4, 2, 2)
    arr[0, :, 0, 0] = np.array([-1.0, -2.0, np.nan, np.nan])
    assert calc_med_first_diffs(arr)[0, 0] == -1
@pytest.mark.skip("Used for local testing")
def test_sigma_clip():
    hdul = fits.open('lrs_TSOjump_sigmaclip5_00_refpix.fits')
    data = hdul['SCI'].data * 4.0
    gdq = hdul['GROUPDQ'].data
    indata = data[:53, :, :, :]
    ingdq = gdq[:53, :, :, :]
    read_noise = np.ones(shape=(indata.shape[2], indata.shape[3]), dtype=np.float32) * 5.9 * 4.0
    hdul.close()
    gdq, row_below_gdq, row_above_gdq, total_primary_crs, stddev = find_crs(indata, ingdq, read_noise, 3,
             4, 5, 1,
             False, 1000,
             10, DQFLAGS,
             after_jump_flag_e1=0.0,
             after_jump_flag_n1=0,
             after_jump_flag_e2=0.0,
             after_jump_flag_n2=0,
             copy_arrs=True, minimum_groups=3, minimum_sigclip_groups=50,)
    fits.writeto("outgdq.fits", gdq, overwrite=True)
@pytest.mark.skip("Used for local testing")
def test_first_grp_flag_issue():
    nints = 8
    nrows = 2
    ncols = 2
    ngroups = 10
    readnoise = 2
    data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    data = np.random.normal(0, readnoise, size=(nints, ngroups, nrows, ncols))
    read_noise = np.full((nrows, ncols), readnoise, dtype=np.float32)
    gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)

    gdq[:, 0, :, :] = DQFLAGS['DO_NOT_USE']
    gdq[1:, 1, :, :] = DQFLAGS['DO_NOT_USE']
    gdq[:, -1, :, :] = DQFLAGS['DO_NOT_USE']
    gdq, row_below_gdq, row_above_gdq, total_total_crs, stddev = \
        find_crs(data, gdq, read_noise, 3, 4, 5, 1, False, 1000, 10, DQFLAGS,
                 after_jump_flag_e1=0.0, after_jump_flag_n1=0,
                 after_jump_flag_e2=0.0, after_jump_flag_n2=0,
                 copy_arrs=True, minimum_groups=3, minimum_sigclip_groups=50)
    fits.writeto("outgdq.fits",gdq, overwrite=True)
@pytest.mark.skip("Used for local testing")
def test_5grp_TSO():
    nints=20
    nrows = 2
    ncols = 2
    ngroups = 5
    readnoise = 25
    data = np.random.normal(0, 0.1 * readnoise, size=(nints, ngroups, nrows, ncols))
    read_noise = np.full((nrows, ncols), readnoise, dtype=np.float32)
    gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)
    np.expand_dims(gdq, axis=0)
    np.expand_dims(data, axis=0)
    gdq[:, 0, :, :] = DQFLAGS['DO_NOT_USE']
    gdq[:, -1, :, :] = DQFLAGS['DO_NOT_USE']
    data[0, :, 0, 0] = [21500, 37600, 52082, 65068, 58627]
    data[0, :, 0, 1] = [21500, 37600, 52082, 65068, 58627]
    gdq, row_below_gdq, row_above_gdq, total_primary_crs, stddev = \
        find_crs(data, gdq, read_noise, 3, 4, 5, 1, False, 1000, 10, DQFLAGS,
                 after_jump_flag_e1=0.0, after_jump_flag_n1=0,
                 after_jump_flag_e2=0.0, after_jump_flag_n2=0,
                 copy_arrs=True, minimum_groups=3, minimum_sigclip_groups=5000)
    fits.writeto("new_gdq.fits", gdq, overwrite=True)
@pytest.mark.skip("Used for local testing")
def test_5grp_realTSO():
    hdul = fits.open("obs2508_cutout_jump.fits")
    gdq = hdul['groupdq'].data
    data = hdul['sci'].data
    readnoise = 25
    read_noise = np.full((3, 3), readnoise, dtype=np.float32)

    gdq, row_below_gdq, row_above_gdq, total_total_crs, stddev = \
        find_crs(data, gdq, read_noise, 3, 4, 5, 1, False, 1000, 10, DQFLAGS,
                 after_jump_flag_e1=0.0, after_jump_flag_n1=0,
                 after_jump_flag_e2=0.0, after_jump_flag_n2=0,
                 copy_arrs=True, minimum_groups=3, minimum_sigclip_groups=15000)
    fits.writeto("new_gdq_cutout.fits", gdq, overwrite=True)



import copy
import logging

import numpy as np
from scipy import ndimage

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def flag_saturated_pixels(
    data, gdq, pdq, sat_thresh, sat_dq, atod_limit, dqflags, n_pix_grow_sat=1, zframe=None, read_pattern=None
):
    """
    Short Summary
    -------------
    Apply flagging for saturation based on threshold values stored in the
    saturation reference file data `sat_thresh` and A/D floor based on testing
    for 0 DN values. For A/D floor flagged groups, the DO_NOT_USE flag is also
    set.

    Parameters
    ----------
    data : float, 4D array
        science array

    gdq : int, 4D array
        group dq array

    pdq : int, 2D array
        pixelg dq array

    sat_thresh : `np.array`
        Pixel-wise threshold for saturation, same shape `data`

    sat_dq : `np.array`
        data quality flags associated with `sat_thresh`

    atod_limit : int
        hard DN limit of 16-bit A-to-D converter

    dqflags : dict
        A dictionary with at least the following keywords:
        DO_NOT_USE, SATURATED, AD_FLOOR, NO_SAT_CHECK

    n_pix_grow_sat : int
        Number of pixels that each flagged saturated pixel should be 'grown',
        to account for charge spilling. Default is 1.

    zframe : float, 3D array
        The ZEROFRAME.

    read_pattern : List[List[float or int]] or None
        The times or indices of the frames composing each group.


    Returns
    -------
    gdq : int, 4D array
        updated group dq array

    pdq : int, 2D array
        updated pixel dq array
    """
    nints, ngroups, nrows, ncols = data.shape
    dnu = int(dqflags["DO_NOT_USE"])
    saturated = int(dqflags["SATURATED"])
    ad_floor = int(dqflags["AD_FLOOR"])
    no_sat_check = int(dqflags["NO_SAT_CHECK"])

    # Identify pixels flagged in reference file as NO_SAT_CHECK,
    no_sat_check_mask = np.bitwise_and(sat_dq, no_sat_check) == no_sat_check

    # Flag pixels in the saturation threshold array with NO_SAT_CHECK,
    # and add them to to no_sat_check_mask.
    sat_dq[np.isnan(sat_thresh)] |= no_sat_check
    no_sat_check_mask |= np.isnan(sat_thresh)

    # Set the saturation check threshold above the A-to-D
    # converter limit so that they don't get flagged as saturated, for
    # pixels in the no_sat_check_mask.
    sat_thresh[no_sat_check_mask] = atod_limit + 1

    for ints in range(nints):
        # Work forward through the groups for initial pass at saturation
        for group in range(ngroups):
            plane = data[ints, group, :, :]

            flagarray, flaglowarray = plane_saturation(plane, sat_thresh, dqflags)

            # for saturation, the flag is set in the current plane
            # and all following planes.
            np.bitwise_or(gdq[ints, group:, :, :], flagarray, gdq[ints, group:, :, :])

            # for A/D floor, the flag is only set of the current plane
            np.bitwise_or(gdq[ints, group, :, :], flaglowarray, gdq[ints, group, :, :])

            del flagarray
            del flaglowarray

            # now, flag any pixels that border saturated pixels (not A/D floor pix)
            if n_pix_grow_sat > 0:
                gdq_slice = copy.copy(gdq[ints, group, :, :]).astype(int)

                gdq[ints, group, :, :] = adjacent_pixels(gdq_slice, saturated, n_pix_grow_sat)

        # Work backward through the groups for a second pass at saturation
        # This is to flag things that actually saturated in prior groups but
        # were not obvious because of group averaging
        for group in range(ngroups-2, -1, -1):
            plane = data[ints, group, :, :]
            thisdq = gdq[ints, group, :, :]
            nextdq = gdq[ints, group + 1, :, :]

            # Determine the dilution factor due to group averaging
            if read_pattern is not None:
                # Single value dilution factor for this group
                dilution_factor = np.mean(read_pattern[group]) / read_pattern[group][-1]
                # Broadcast to array size
                dilution_factor = np.where(no_sat_check_mask, 1, dilution_factor)
            else:
                dilution_factor = 1

            # Find where this plane looks like it might saturate given the dilution factor
            flagarray, _ = plane_saturation(plane, sat_thresh * dilution_factor, dqflags)

            # Find the overlap of where this plane looks like it might saturate, was not currently
            # flagged as saturation or DO_NOT_USE, and the next group had saturation flagged.
            indx = np.where((np.bitwise_and(flagarray, saturated) != 0) & \
                            (np.bitwise_and(thisdq, saturated) == 0) & \
                            (np.bitwise_and(thisdq, dnu) == 0) & \
                            (np.bitwise_and(nextdq, saturated) != 0))

            # Reset flag array to only pixels passing this gauntlet
            flagarray[:] = 0
            flagarray[indx] = dnu
            
            # Grow the newly-flagged saturating pixels
            if n_pix_grow_sat > 0:
                flagarray = adjacent_pixels(flagarray, dnu, n_pix_grow_sat)

            # Add them to the gdq array
            np.bitwise_or(gdq[ints, group, :, :], flagarray, gdq[ints, group, :, :])

        # Add an additional pass to look for things saturating in the second group
        # that can be particularly tricky to identify
        if ((read_pattern is not None) & (ngroups > 2)):
            dq2 = gdq[ints, 1, :, :]
            dq3 = gdq[ints, 2, :, :]
            
            # Identify groups which we wouldn't expect to saturate by the third group,
            # on the basis of the first group
            scigp1 = data[ints, 0, :, :]
            mask = scigp1 / np.mean(read_pattern[0]) * read_pattern[2][-1] < sat_thresh

            # Identify groups with suspiciously large values in the second group
            # In the limit of groups with just nframe this just checks if second group
            # is over the regular saturation limit.
            scigp2 = data[ints, 1, :, :]
            mask &= scigp2 > sat_thresh / len(read_pattern[1])

            # Identify groups that are saturated in the third group but not yet flagged in the second
            gp3mask = np.where((np.bitwise_and(dq3, saturated) != 0) & \
                               (np.bitwise_and(dq2, saturated) == 0), True, False)
            mask &= gp3mask

            # Flag the 2nd group for the pixels passing that gauntlet
            flagarray = np.zeros_like(mask,dtype='uint8')
            flagarray[mask] = dnu

            # Add them to the gdq array
            np.bitwise_or(gdq[ints, 1, :, :], flagarray, gdq[ints, 1, :, :])
            

        # Check ZEROFRAME.
        if zframe is not None:
            plane = zframe[ints, :, :]
            flagarray, flaglowarray = plane_saturation(plane, sat_thresh, dqflags)
            zdq = flagarray | flaglowarray
            if n_pix_grow_sat > 0:
                zdq = adjacent_pixels(zdq, saturated, n_pix_grow_sat)
            plane[zdq != 0] = 0.0
            zframe[ints] = plane

    n_sat = np.any(np.any(np.bitwise_and(gdq, saturated), axis=0), axis=0).sum()
    log.info("Detected %i saturated pixels", n_sat)
    n_floor = np.any(np.any(np.bitwise_and(gdq, ad_floor), axis=0), axis=0).sum()
    log.info("Detected %i A/D floor pixels", n_floor)

    pdq = np.bitwise_or(pdq, sat_dq)

    return gdq, pdq, zframe


def adjacent_pixels(plane_gdq, saturated, n_pix_grow_sat):
    """
    plane_gdq : ndarray
        The data quality flags of the current.

    saturated : uint8
        The saturation flag.

    n_pix_grow_sat : int
        Number of pixels that each flagged saturated pixel should be 'grown',
        to account for charge spilling. Default is 1.

    Return
    ------
    sat_pix : ndarray
        The saturated pixels in the current plane.
    """
    cgdq = plane_gdq.copy()
    only_sat = np.bitwise_and(plane_gdq, saturated).astype(np.uint8)
    box_dim = (n_pix_grow_sat * 2) + 1
    struct = np.ones((box_dim, box_dim)).astype(bool)
    dialated = ndimage.binary_dilation(only_sat, structure=struct).astype(only_sat.dtype)
    return np.bitwise_or(cgdq, (dialated * saturated))


def plane_saturation(plane, sat_thresh, dqflags):
    """
    plane : ndarray, 2D float
        The plane to check for saturation and A/D floor.

    sat_thresh : `np.array`
        Pixel-wise threshold for saturation, same shape `data`.

    dims : tuple
        The dimensions of the data array.

    dqflags : dict
        A dictionary with at least the following keywords:
        DO_NOT_USE, SATURATED, AD_FLOOR, NO_SAT_CHECK
    """
    donotuse = dqflags["DO_NOT_USE"]
    saturated = dqflags["SATURATED"]
    ad_floor = dqflags["AD_FLOOR"]

    flagarray = np.zeros(plane.shape, dtype=np.uint32)
    flaglowarray = np.zeros(plane.shape, dtype=np.uint32)

    # Update the 4D gdq array with the saturation flag.
    # check for saturation
    flagarray[:, :] = np.where(plane[:, :] >= sat_thresh, saturated, 0)

    # check for A/D floor
    flaglowarray[:, :] = np.where(plane[:, :] <= 0, ad_floor | donotuse, 0)

    return flagarray, flaglowarray



#! /usr/bin/env python
#
#  ramp_fit.py - calculate weighted mean of slope, based on Massimo
#                Robberto's "On the Optimal Strategy to fit MULTIACCUM
#                ramps in the presence of cosmic rays."
#                (JWST-STScI-0001490,SM-12; 07/25/08).   The derivation
#                is a generalization for >1 cosmic rays, calculating
#                the slope and variance of the slope for each section
#                of the ramp (in between cosmic rays). The intervals are
#                determined from the input data quality arrays.
#
# Note:
# In this module, comments on the 'first group','second group', etc are
#    1-based, unless noted otherwise.

import logging

import numpy as np
from astropy import units as u

from . import (
    gls_fit,    # used only if algorithm is "GLS"
    likely_fit, # used only if algorithm is "LIKELY"
    ols_fit,    # used only if algorithm is "OLS"
    ramp_fit_class,
)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

BUFSIZE = 1024 * 300000  # 300Mb cache size for data section


def create_ramp_fit_class(model, algorithm, dqflags=None, suppress_one_group=False):
    """
    Create an internal ramp fit class from a data model.

    Parameters
    ----------
    model : data model
        input data model, assumed to be of type RampModel

    dqflags : dict
        The data quality flags needed for ramp fitting.

    suppress_one_group : bool
        Find ramps with only one good group and treat it like it has zero good
        groups.

    Return
    ------
    ramp_data : ramp_fit_class.RampData
        The internal ramp class.
    """
    ramp_data = ramp_fit_class.RampData()

    if not hasattr(model, 'average_dark_current'):
        dark_current_array = np.zeros_like(model.pixeldq)
    else:
        dark_current_array = model.average_dark_current

    orig_gdq = None
    if algorithm.upper() == "OLS_C":
        wh_chargeloss = np.where(np.bitwise_and(model.groupdq.astype(np.uint32), dqflags['CHARGELOSS']))
        if len(wh_chargeloss[0]) > 0:
            orig_gdq = model.groupdq.copy()
        del wh_chargeloss

    if isinstance(model.data, u.Quantity):
        ramp_data.set_arrays(model.data.value, model.err.value, model.groupdq,
                             model.pixeldq, dark_current_array)
    else:
        ramp_data.set_arrays(
            model.data,
            model.err,
            model.groupdq,
            model.pixeldq,
            dark_current_array,
            orig_gdq)

    # Attribute may not be supported by all pipelines.  Default is NoneType.
    drop_frames1 = model.meta.exposure.drop_frames1 if hasattr(model, "drop_frames1") else None
    ramp_data.set_meta(
        name=model.meta.instrument.name,
        frame_time=model.meta.exposure.frame_time,
        group_time=model.meta.exposure.group_time,
        groupgap=model.meta.exposure.groupgap,
        nframes=model.meta.exposure.nframes,
        drop_frames1=drop_frames1,
    )

    if "zero_frame" in model.meta.exposure and model.meta.exposure.zero_frame:
        ramp_data.zeroframe = model.zeroframe

    ramp_data.algorithm = algorithm

    if hasattr(model.meta.exposure, "read_pattern"):
        ramp_data.read_pattern = [list(reads) for reads in model.meta.exposure.read_pattern]

    ramp_data.set_dqflags(dqflags)
    ramp_data.start_row = 0
    ramp_data.num_rows = ramp_data.data.shape[2]

    ramp_data.suppress_one_group_ramps = suppress_one_group

    return ramp_data


def ramp_fit(
    model,
    buffsize,
    save_opt,
    readnoise_2d,
    gain_2d,
    algorithm,
    weighting,
    max_cores,
    dqflags,
    suppress_one_group=False,
):
    """
    Calculate the count rate for each pixel in all data cube sections and all
    integrations, equal to the slope for all sections (intervals between
    cosmic rays) of the pixel's ramp divided by the effective integration time.
    The weighting parameter must currently be set to 'optim', to use the optimal
    weighting (paper by Fixsen, ref. TBA) will be used in the fitting; this is
    currently the only supported weighting scheme.

    Parameters
    ----------
    model : data model
        input data model, assumed to be of type RampModel

    buffsize : int
        size of data section (buffer) in bytes

    save_opt : bool
       calculate optional fitting results

    readnoise_2d : ndarray
        2-D array readnoise for all pixels

    gain_2d : ndarray
        2-D array gain for all pixels

    algorithm : str
        'OLS' specifies that ordinary least squares should be used;
        'GLS' specifies that generalized least squares should be used.
        'LIKELY' specifies that maximum likelihood should be used.

    weighting : str
        'optimal' specifies that optimal weighting should be used;
         currently the only weighting supported.

    max_cores : str
        Number of cores to use for multiprocessing. If set to 'none' (the
        default), then no multiprocessing will be done. The other allowable
        values are 'quarter', 'half', and 'all'. This is the fraction of cores
        to use for multi-proc. The total number of cores includes the SMT cores
        (Hyper Threading for Intel).

    dqflags : dict
        A dictionary with at least the following keywords:
        DO_NOT_USE, SATURATED, JUMP_DET, NO_GAIN_VALUE, UNRELIABLE_SLOPE

    suppress_one_group : bool
        Find ramps with only one good group and treat it like it has zero good
        groups.

    Returns
    -------
    image_info : tuple
        The tuple of computed ramp fitting arrays.

    integ_info : tuple
        The tuple of computed integration fitting arrays.

    opt_info : tuple
        The tuple of computed optional results arrays for fitting.

    gls_opt_model : GLS_RampFitModel object or None (Unused for now)
        Object containing optional GLS-specific ramp fitting data for the
        exposure
    """
    if suppress_one_group and model.data.shape[1] == 1:
        # One group ramp suppression should only be done on data with
        # ramps having more than one group.
        suppress_one_group = False

    # Create an instance of the internal ramp class, using only values needed
    # for ramp fitting from the to remove further ramp fitting dependence on
    # data models.
    ramp_data = create_ramp_fit_class(model, algorithm, dqflags, suppress_one_group)

    return ramp_fit_data(
        ramp_data, buffsize, save_opt, readnoise_2d, gain_2d, algorithm, weighting, max_cores, dqflags
    )


def ramp_fit_data(
    ramp_data, buffsize, save_opt, readnoise_2d, gain_2d, algorithm, weighting, max_cores, dqflags
):
    """
    This function begins the ramp fit computation after the creation of the
    RampData class.  It determines the proper path for computation to take
    depending on the choice of ramp fitting algorithms (which is only ordinary
    least squares right now) and the choice of single or muliprocessing.


    ramp_data : RampData
        Input data necessary for computing ramp fitting.

    buffsize : int
        size of data section (buffer) in bytes

    save_opt : bool
       calculate optional fitting results

    readnoise_2d : ndarray
        2-D array readnoise for all pixels

    gain_2d : ndarray
        2-D array gain for all pixels

    algorithm : str
        'OLS' specifies that ordinary least squares should be used;
        'GLS' specifies that generalized least squares should be used.
        'LIKELY' specifies that maximum likelihood should be used.

    weighting : str
        'optimal' specifies that optimal weighting should be used;
         currently the only weighting supported.

    max_cores : str
        Number of cores to use for multiprocessing. If set to 'none' (the
        default), then no multiprocessing will be done. The other allowable
        values are 'quarter', 'half', and 'all'. This is the fraction of cores
        to use for multi-proc. The total number of cores includes the SMT cores
        (Hyper Threading for Intel).

    dqflags : dict
        A dictionary with at least the following keywords:
        DO_NOT_USE, SATURATED, JUMP_DET, NO_GAIN_VALUE, UNRELIABLE_SLOPE

    Returns
    -------
    image_info : tuple
        The tuple of computed ramp fitting arrays.

    integ_info : tuple
        The tuple of computed integration fitting arrays.

    opt_info : tuple
        The tuple of computed optional results arrays for fitting.

    gls_opt_model : GLS_RampFitModel object or None (Unused for now)
        Object containing optional GLS-specific ramp fitting data for the
        exposure
    """
    # For the LIKELY algorithm, due to the jump detection portion of the code
    # a minimum of a four group ramp is needed.
    ngroups = ramp_data.data.shape[1]
    if algorithm.upper() == "LIKELY" and ngroups < likely_fit.LIKELY_MIN_NGROUPS:
        log.info(f"When selecting the LIKELY ramp fitting algorithm the"
                  " ngroups needs to be a minimum of {likely_fit.LIKELY_MIN_NGROUPS},"
                  " but ngroups = {ngroups}.  Due to this, the ramp fitting algorithm"
                  " is being changed to OLS_C")
        algorithm = "OLS_C"

    if algorithm.upper() == "OLS_C":
        ramp_data.run_c_code = True
        
    if algorithm.upper() == "GLS":
        image_info, integ_info, gls_opt_info = gls_fit.gls_ramp_fit(
            ramp_data, buffsize, save_opt, readnoise_2d, gain_2d, max_cores
        )
        opt_info = None
    elif algorithm.upper() == "LIKELY" and ngroups >= likely_fit.LIKELY_MIN_NGROUPS:
        image_info, integ_info, opt_info = likely_fit.likely_ramp_fit(
            ramp_data, readnoise_2d, gain_2d
        )
        gls_opt_info = None
    else:
        # Default to OLS.
        # Get readnoise array for calculation of variance of noiseless ramps, and
        #   gain array in case optimal weighting is to be done
        nframes = ramp_data.nframes
        readnoise_2d *= gain_2d / np.sqrt(2.0 * nframes)

        # Suppress one group ramps, if desired.
        if ramp_data.suppress_one_group_ramps:
            suppress_one_good_group_ramps(ramp_data)

        # Compute ramp fitting using ordinary least squares.
        image_info, integ_info, opt_info = ols_fit.ols_ramp_fit_multi(
            ramp_data, buffsize, save_opt, readnoise_2d, gain_2d, weighting, max_cores
        )
        gls_opt_info = None

    return image_info, integ_info, opt_info, gls_opt_info


def suppress_one_good_group_ramps(ramp_data):
    """
    Finds one group ramps in each integration and suppresses them, i.e. turns
    them into zero group ramps.

    Parameter
    ---------
    ramp_data : RampData
        input data model, assumed to be of type RampModel
    """
    dq = ramp_data.groupdq
    nints, ngroups, nrows, ncols = dq.shape
    dnu_flag = ramp_data.flags_do_not_use

    for integ in range(nints):
        # In the current integration find ramps with only one group.
        intdq = dq[integ, :, :, :]
        good_groups = np.zeros(intdq.shape, dtype=int)
        good_groups[intdq == 0] = 1
        ngood_groups = good_groups.sum(axis=0)
        wh_one = np.where(ngood_groups == 1)

        # Suppress the ramps with only one good group by flagging
        # all groups in the ramp as DO_NOT_USE.
        wh1_rows = wh_one[0]
        wh1_cols = wh_one[1]
        for n in range(len(wh1_rows)):
            row = wh1_rows[n]
            col = wh1_cols[n]
            # Find ramps that have good 0th group, but the rest of the
            # ramp flagged.
            good_index = np.where(ramp_data.groupdq[integ, :, row, col] == 0)
            if ramp_data.groupdq[integ, good_index, row, col] == 0:
                ramp_data.groupdq[integ, :, row, col] = np.bitwise_or(
                    ramp_data.groupdq[integ, :, row, col], dnu_flag
                )



import logging
import multiprocessing
import time
import warnings

import numpy as np
import cv2 as cv
import astropy.stats as stats

from astropy.convolution import Ring2DKernel
from astropy.convolution import convolve

from . import constants
from . import twopoint_difference as twopt

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def detect_jumps(
    frames_per_group,
    indata,
    gdq,
    pdq,
    inerr,
    gain_2d,
    readnoise_2d,
    rejection_thresh,
    three_grp_thresh,
    four_grp_thresh,
    max_cores,
    max_jump_to_flag_neighbors,
    min_jump_to_flag_neighbors,
    flag_4_neighbors,
    dqflags,
    after_jump_flag_dn1=0.0,
    after_jump_flag_n1=0,
    after_jump_flag_dn2=0.0,
    after_jump_flag_n2=0,
    min_sat_area=1,
    min_jump_area=5,
    expand_factor=2.0,
    use_ellipses=False,
    sat_required_snowball=True,
    expand_large_events=False,
    sat_expand=2,
    min_sat_radius_extend=2.5,
    find_showers=False,
    edge_size=25,
    extend_snr_threshold=1.2,
    extend_min_area=90,
    extend_inner_radius=1,
    extend_outer_radius=2.6,
    extend_ellipse_expand_ratio=1.2,
    grps_masked_after_shower=5,
    max_extended_radius=200,
    minimum_groups=3,
    minimum_sigclip_groups=100,
    only_use_ints=True,
    min_diffs_single_pass=10,
    mask_persist_grps_next_int=True,
    persist_grps_flagged=25,
    max_shower_amplitude=12
):
    """
    This is the high-level controlling routine for the jump detection process.
    It loads and sets the various input data and parameters needed by each of
    the individual detection methods and then calls the detection methods in
    turn.

    Note that the detection methods are currently set up on the assumption
    that the input science and error data arrays will be in units of
    electrons, hence this routine scales those input arrays by the detector
    gain. The methods assume that the read noise values will be in units
    of DN.

    The gain is applied to the science data and error arrays using the
    appropriate instrument- and detector-dependent values for each pixel of an
    image.  Also, a 2-dimensional read noise array with appropriate values for
    each pixel is passed to the detection methods.

    Parameters
    ----------
    frames_per_group : int
        number of frames per group

    indata : float, 4D array
        science array

    gdq : int, 4D array
        group dq array

    pdq : int, 2D array
        pixelg dq array

    inerr : float, 4D array
        error array

    gain_2d : float, 2D array
        gain for all pixels

    readnoise_2d : float, 2D array
        readnoise for all pixels

    rejection_thresh : float
        The 'normal' cosmic ray sigma rejection threshold for ramps with more
        than 4 groups

    three_grp_thresh : float
        cosmic ray sigma rejection threshold for ramps having 3 groups

    four_grp_thresh : float
        cosmic ray sigma rejection threshold for ramps having 4 groups

    max_cores : str
        Maximum number of cores to use for multiprocessing. Available choices
        are 'none' (which will create one process), 'quarter', 'half', 'all'
        (of available cpu cores).

    max_jump_to_flag_neighbors : float
        value in units of sigma that sets the upper limit for flagging of
        neighbors. Any jump above this cutoff will not have its neighbors
        flagged.

    min_jump_to_flag_neighbors : float
        value in units of sigma that sets the lower limit for flagging of
        neighbors (marginal detections). Any primary jump below this value will
        not have its neighbors flagged.

    flag_4_neighbors : bool
        if set to True (default is True), it will cause the four perpendicular
        neighbors of all detected jumps to also be flagged as a jump.

    dqflags : dict
        A dictionary with at least the following keywords:
        DO_NOT_USE, SATURATED, JUMP_DET, NO_GAIN_VALUE, GOOD

    after_jump_flag_dn1 : float
        Jumps with amplitudes above the specified DN value will have subsequent
        groups flagged with the number determined by the after_jump_flag_n1

    after_jump_flag_n1 : int
        Gives the number of groups to flag after jumps with DN values above that
        given by after_jump_flag_dn1

    after_jump_flag_dn2 : float
        Jumps with amplitudes above the specified DN value will have subsequent
        groups flagged with the number determined by the after_jump_flag_n2

    after_jump_flag_n2 : int
        Gives the number of groups to flag after jumps with DN values above that
        given by after_jump_flag_dn2

    min_sat_area : float
        The minimum area of saturated pixels at the center of a snowball. Only
        contours with area above the minimum will create snowballs.

    min_jump_area : float
        The minimum contour area to trigger the creation of enclosing ellipses
        or circles.

    expand_factor : float
        The factor that is used to increase the size of the enclosing
        circle/ellipse jump flagged pixels.

    use_ellipses : deprecated

    sat_required_snowball : bool
        If true there must be a saturation circle within the radius of the jump
        circle to trigger the creation of a snowball. All true snowballs appear
        to have at least one saturated pixel.

    edge_size : int
        The distance from the edge of the detector where saturated cores are not
        required for snowball detection

    expand_large_events : bool
        When True this triggers the flagging of snowballs for NIR detectors.

    sat_expand : int
        The number of pixels to expand the saturated core of detected snowballs

    find_showers : boolean
        Turns on the flagging of the faint extended emission of MIRI showers

    extend_snr_threshold : float
        The SNR minimum for the detection of faint extended showers in MIRI

    extend_min_area : float
        The required minimum area of extended emission after convolution for the
        detection of showers in MIRI

    extend_inner_radius : float
        The inner radius of the Ring2DKernal that is used for the detection of
        extended emission in showers

    extend_outer_radius : float
        The outer radius of the Ring2DKernal that is used for the detection of
        extended emission in showers

    extend_ellipse_expand_ratio : float
        Multiplicative factor to expand the radius of the ellipse fit to the
        detected extended emission in MIRI showers

    grps_masked_after_shower : int
        Number of groups after detected extended emission to flag as a jump for
        MIRI showers

    max_extended_radius : int
        The maximum radius for any extension of saturation or jump

    min_sat_radius_extend : float
        The minimum radius of the saturated core of a snowball for the core to
        be extended
    minimum_groups : int
       The minimum number of groups for jump detection
    minimum_sigclip_groups : int
       The minimum number of groups required to use sigma clipping to find outliers.
    only_use_ints : boolean
       In sigma clipping, if True only differences between integrations are compared. If False,
       then all differences are processed at once.
    min_diffs_single_pass : int
       The minimum number of groups to switch to flagging all outliers in a single pass.
    max_shower_amplitude : float
       The maximum possible amplitude for flagged MIRI showers in DN/group

    Returns
    -------
    gdq : int, 4D array
        updated group dq array

    pdq : int, 2D array
        updated pixel dq array
    """
    constants.update_dqflags(dqflags)  # populate dq flags
    sat_flag = dqflags["SATURATED"]
    jump_flag = dqflags["JUMP_DET"]
    number_extended_events = 0
    # Flag the pixeldq where the gain is <=0 or NaN so they will be ignored
    wh_g = np.where(gain_2d <= 0.0)
    if len(wh_g[0] > 0):
        pdq[wh_g] = np.bitwise_or(pdq[wh_g], dqflags["NO_GAIN_VALUE"])
        pdq[wh_g] = np.bitwise_or(pdq[wh_g], dqflags["DO_NOT_USE"])

    wh_g = np.where(np.isnan(gain_2d))
    if len(wh_g[0] > 0):
        pdq[wh_g] = np.bitwise_or(pdq[wh_g], dqflags["NO_GAIN_VALUE"])
        pdq[wh_g] = np.bitwise_or(pdq[wh_g], dqflags["DO_NOT_USE"])

    # Apply gain to the SCI, ERR, and readnoise arrays so they're in units
    # of electrons
    data = indata * gain_2d
    err = inerr * gain_2d
    readnoise_2d *= gain_2d
    # also apply to the after_jump thresholds
    after_jump_flag_e1 = after_jump_flag_dn1 * np.nanmedian(gain_2d)
    after_jump_flag_e2 = after_jump_flag_dn2 * np.nanmedian(gain_2d)
    # Apply the 2-point difference method as a first pass
    log.info("Executing two-point difference method")
    start = time.time()

    # Set parameters of input data shape
    n_rows = data.shape[-2]
    n_cols = data.shape[-1]
    n_groups = data.shape[1]
    n_ints = data.shape[0]

    row_above_gdq = np.zeros((n_ints, n_groups, n_cols), dtype=np.uint8)
    previous_row_above_gdq = np.zeros((n_ints, n_groups, n_cols), dtype=np.uint8)
    row_below_gdq = np.zeros((n_ints, n_groups, n_cols), dtype=np.uint8)

    # figure out how many slices to make based on 'max_cores'
    max_available = multiprocessing.cpu_count()
    n_slices = calc_num_slices(n_rows, max_cores, max_available)
    if n_slices == 1:
        gdq, row_below_dq, row_above_dq, total_primary_crs, stddev = twopt.find_crs(
            data,
            gdq,
            readnoise_2d,
            rejection_thresh,
            three_grp_thresh,
            four_grp_thresh,
            frames_per_group,
            flag_4_neighbors,
            max_jump_to_flag_neighbors,
            min_jump_to_flag_neighbors,
            dqflags,
            after_jump_flag_e1=after_jump_flag_e1,
            after_jump_flag_n1=after_jump_flag_n1,
            after_jump_flag_e2=after_jump_flag_e2,
            after_jump_flag_n2=after_jump_flag_n2,
            copy_arrs=False,
            minimum_groups=3,
            minimum_sigclip_groups=minimum_sigclip_groups,
            only_use_ints=only_use_ints,
            min_diffs_single_pass=min_diffs_single_pass,
        )
        # remove redundant bits in pixels that have jump flagged but were
        # already flagged as do_not_use or saturated.
        gdq[gdq == np.bitwise_or(dqflags['DO_NOT_USE'], dqflags['JUMP_DET'])] = \
            dqflags['DO_NOT_USE']
        gdq[gdq == np.bitwise_or(dqflags['SATURATED'], dqflags['JUMP_DET'])] = \
            dqflags['SATURATED']

    else:
        yinc = int(n_rows // n_slices)
        slices = []
        # Slice up data, gdq, readnoise_2d into slices
        # Each element of slices is a tuple of
        # (data, gdq, readnoise_2d, rejection_thresh, three_grp_thresh,
        #  four_grp_thresh, nframes)

        # must copy arrays here, find_crs will make copies but if slices
        # are being passed in for multiprocessing then the original gdq will be
        # modified unless copied beforehand.
        gdq = gdq.copy()
        data = data.copy()
        copy_arrs = False  # we don't need to copy arrays again in find_crs
        for i in range(n_slices - 1):
            slices.insert(
                i,
                (
                    data[:, :, i * yinc: (i + 1) * yinc, :],
                    gdq[:, :, i * yinc: (i + 1) * yinc, :].copy(),
                    readnoise_2d[i * yinc: (i + 1) * yinc, :],
                    rejection_thresh,
                    three_grp_thresh,
                    four_grp_thresh,
                    frames_per_group,
                    flag_4_neighbors,
                    max_jump_to_flag_neighbors,
                    min_jump_to_flag_neighbors,
                    dqflags,
                    after_jump_flag_e1,
                    after_jump_flag_n1,
                    after_jump_flag_e2,
                    after_jump_flag_n2,
                    copy_arrs,
                    minimum_groups,
                    minimum_sigclip_groups,
                    only_use_ints,
                    min_diffs_single_pass,
                ),
            )

        # last slice get the rest
        slices.insert(
            n_slices - 1,
            (
                data[:, :, (n_slices - 1) * yinc: n_rows, :],
                gdq[:, :, (n_slices - 1) * yinc: n_rows, :].copy(),
                readnoise_2d[(n_slices - 1) * yinc: n_rows, :],
                rejection_thresh,
                three_grp_thresh,
                four_grp_thresh,
                frames_per_group,
                flag_4_neighbors,
                max_jump_to_flag_neighbors,
                min_jump_to_flag_neighbors,
                dqflags,
                after_jump_flag_e1,
                after_jump_flag_n1,
                after_jump_flag_e2,
                after_jump_flag_n2,
                copy_arrs,
                minimum_groups,
                minimum_sigclip_groups,
                only_use_ints,
                min_diffs_single_pass,
            ),
        )
        log.info("Creating %d processes for jump detection ", n_slices)
        ctx = multiprocessing.get_context("forkserver")
        pool = ctx.Pool(processes=n_slices)
        ######### JUST FOR DEBUGGING #########################
        # pool = ctx.Pool(processes=1)
        # Starts each slice in its own process. Starmap allows more than one
        # parameter to be passed.
        real_result = pool.starmap(twopt.find_crs, slices)
        pool.close()
        pool.join()
        k = 0

        # Reconstruct gdq, the row_above_gdq, and the row_below_gdq from the
        # slice result
        total_primary_crs = 0
        ngrps = gdq.shape[1]
        nrows = gdq.shape[2]
        ncols = gdq.shape[3]
        if only_use_ints:
            stddev = np.zeros((ngrps - 1, nrows, ncols), dtype=np.float32)
        else:
            stddev = np.zeros((nrows, ncols), dtype=np.float32)
        for resultslice in real_result:
            if len(real_result) == k + 1:  # last result
                gdq[:, :, k * yinc: n_rows, :] = resultslice[0]
                if only_use_ints:
                    stddev[:, k * yinc: n_rows, :] = resultslice[4]
                else:
                    stddev[k * yinc: n_rows, :] = resultslice[4]
            else:
                gdq[:, :, k * yinc: (k + 1) * yinc, :] = resultslice[0]
                if only_use_ints:
                    stddev[:, k * yinc: (k + 1) * yinc, :] = resultslice[4]
                else:
                    stddev[k * yinc : (k + 1) * yinc, :] = resultslice[4]
            row_below_gdq[:, :, :] = resultslice[1]
            row_above_gdq[:, :, :] = resultslice[2]
            total_primary_crs += resultslice[3]
            if k != 0:
                # For all but the first slice, flag any CR neighbors in the top
                # row of the previous slice and flag any neighbors in the
                # bottom row of this slice saved from the top of the previous
                # slice
                gdq[:, :, k * yinc - 1, :] = np.bitwise_or(gdq[:, :, k * yinc - 1, :], row_below_gdq[:, :, :])
                gdq[:, :, k * yinc, :] = np.bitwise_or(
                    gdq[:, :, k * yinc, :], previous_row_above_gdq[:, :, :]
                )

            # save the neighbors to be flagged that will be in the next slice
            previous_row_above_gdq = row_above_gdq.copy()
            k += 1
        # remove redundant bits in pixels that have jump flagged but were
        # already flagged as do_not_use or saturated.
        gdq[gdq == np.bitwise_or(dqflags['DO_NOT_USE'], dqflags['JUMP_DET'])] = \
            dqflags['DO_NOT_USE']
        gdq[gdq == np.bitwise_or(dqflags['SATURATED'], dqflags['JUMP_DET'])] = \
            dqflags['SATURATED']

    #  Look for snowballs in near-IR data
    if expand_large_events:
        gdq, total_snowballs = flag_large_events(
            gdq,
            jump_flag,
            sat_flag,
            min_sat_area=min_sat_area,
            min_jump_area=min_jump_area,
            expand_factor=expand_factor,
            sat_required_snowball=sat_required_snowball,
            min_sat_radius_extend=min_sat_radius_extend,
            edge_size=edge_size,
            sat_expand=sat_expand,
            max_extended_radius=max_extended_radius,
            mask_persist_grps_next_int=mask_persist_grps_next_int,
            persist_grps_flagged=persist_grps_flagged,
        )
        log.info("Total snowballs = %i", total_snowballs)
        number_extended_events = total_snowballs

    # Look for showers in mid-IR data
    if find_showers:
        gdq, num_showers = find_faint_extended(
            data,
            gdq,
            pdq,
            readnoise_2d,
            frames_per_group,
            minimum_sigclip_groups,
            dqflags,
            snr_threshold=extend_snr_threshold,
            min_shower_area=extend_min_area,
            inner=extend_inner_radius,
            outer=extend_outer_radius,
            sat_flag=sat_flag,
            jump_flag=jump_flag,
            ellipse_expand=extend_ellipse_expand_ratio,
            num_grps_masked=grps_masked_after_shower,
            max_extended_radius=max_extended_radius,
            max_shower_amplitude=max_shower_amplitude
        )
        log.info("Total showers= %i", num_showers)
        number_extended_events = num_showers

    elapsed = time.time() - start
    log.info("Total elapsed time = %g sec", elapsed)

    # Back out the applied gain to the SCI, ERR, and readnoise arrays so they're
    #    back in units of DN
    data /= gain_2d
    err /= gain_2d
    readnoise_2d /= gain_2d
    # Return the updated data quality arrays
    return gdq, pdq, total_primary_crs, number_extended_events, stddev


def flag_large_events(
    gdq,
    jump_flag,
    sat_flag,
    min_sat_area=1,
    min_jump_area=6,
    expand_factor=2.0,
    sat_required_snowball=True,
    min_sat_radius_extend=2.5,
    sat_expand=2,
    edge_size=25,
    max_extended_radius=200,
    mask_persist_grps_next_int=True,
    persist_grps_flagged=5,
):
    """
    This routine controls the creation of expanded regions that are flagged as
    jumps.

    These events are called snowballs for the NIR. While they are most commonly
    circular, there are elliptical ones. This routine does not handle the
    detection of MIRI showers.

    Parameters
    ----------
    gdq : int, 4D array
        Group dq array
    jump_flag : int
        DQ flag for jump detection.
    sat_flag: int
        DQ flag for saturation
    min_sat_area : int
        The minimum area of saturated pixels within the jump circle to trigger
        the creation of a snowball.
    min_jump_area : int
        The minimum area of a contour to cause the creation of the
        minimum enclosing ellipse or circle.
    expand_factor : float
        The factor that increases the size of the snowball or enclosing ellipse.
    sat_required_snowball : bool
        Require that there is a saturated pixel within the radius of the jump
        circle to trigger the formation of a snowball.
    min_sat_radius_extend : float
        The smallest radius to trigger extension of the saturated core
    sat_expand : int
        The number of pixels to extend the saturated core by
    edge_size : int
        The distance from the edge of the detector where saturation is not
        required for a snowball to be created
    max_extended_radius : int
        The largest radius that a snowball or shower can be extended
    mask_persist_grps_next_int : bool
        The flag to turn on the extension of the flagging of the saturated cores of
        snowballs.
    persist_grps_flagged : int
        How many groups to be flagged when the saturated cores are extended into
        subsequent integrations
    Returns
    -------
    total Snowballs

    """
    log.info("Flagging Snowballs")

    n_showers_grp = []
    total_snowballs = 0
    nints, ngrps, nrows, ncols = gdq.shape
    persist_jumps = np.zeros(shape=(nints, gdq.shape[2], gdq.shape[3]), dtype=np.uint8)
    for integration in range(nints):
        for group in range(1, ngrps):
            current_gdq = gdq[integration, group, :, :]
            current_sat = np.bitwise_and(current_gdq, sat_flag)
            prev_gdq = gdq[integration, group - 1, :, :]
            prev_sat = np.bitwise_and(prev_gdq, sat_flag)
            not_prev_sat = np.logical_not(prev_sat)
            new_sat = current_sat * not_prev_sat
            if group < ngrps - 1:
                next_gdq = gdq[integration, group + 1, :, :]
                next_sat = np.bitwise_and(next_gdq, sat_flag)
                not_current_sat = np.logical_not(current_sat)
                next_new_sat = next_sat * not_current_sat
            next_sat_ellipses = find_ellipses(next_new_sat, sat_flag, min_sat_area)
            sat_ellipses = find_ellipses(new_sat, sat_flag, min_sat_area)
            # find the ellipse parameters for jump regions
            jump_ellipses = find_ellipses(gdq[integration, group, :, :], jump_flag, min_jump_area)
            if sat_required_snowball:
                low_threshold = edge_size
                high_threshold = max(0, nrows - edge_size)
                gdq, snowballs, persist_jumps = make_snowballs(
                    gdq,
                    integration,
                    group,
                    jump_ellipses,
                    sat_ellipses,
                    next_sat_ellipses,
                    low_threshold,
                    high_threshold,
                    min_sat_radius_extend,
                    sat_expand,
                    sat_flag,
                    jump_flag,
                    max_extended_radius,
                    persist_jumps,
                )
            else:
                snowballs = jump_ellipses
            n_showers_grp.append(len(snowballs))
            total_snowballs += len(snowballs)
            gdq, num_events = extend_ellipses(
                gdq,
                integration,
                group,
                snowballs,
                sat_flag,
                jump_flag,
                expansion=expand_factor,
                num_grps_masked=0,
                max_extended_radius=max_extended_radius,
            )

    #  Test to see if the flagging of the saturated cores will be extended into the
    #  subsequent integrations. Persist_jumps contains all the pixels that were saturated
    #  in the cores of snowballs.
    if mask_persist_grps_next_int:
        for intg in range(1, nints):
            if persist_grps_flagged >= 1:
                last_grp_flagged = min(persist_grps_flagged, ngrps)
                gdq[intg, 1:last_grp_flagged, :, :] = np.bitwise_or(gdq[intg, 1:last_grp_flagged, :, :],
                                                                    np.repeat(persist_jumps[intg - 1, np.newaxis, :, :],
                                                                    last_grp_flagged - 1, axis=0))
    return gdq, total_snowballs

def extend_saturation(
    cube, grp, sat_ellipses, sat_flag, jump_flag, min_sat_radius_extend, persist_jumps,
    expansion=2, max_extended_radius=200
):
    ngroups, nrows, ncols = cube.shape
    image = np.zeros(shape=(nrows, ncols, 3), dtype=np.uint8)
    persist_image = np.zeros(shape=(nrows, ncols, 3), dtype=np.uint8)
    outcube = cube.copy()
    for ellipse in sat_ellipses:
        ceny = ellipse[0][0]
        cenx = ellipse[0][1]
        minor_axis = min(ellipse[1][1], ellipse[1][0])
        if minor_axis > min_sat_radius_extend:
            axis1 = ellipse[1][0] + expansion
            axis2 = ellipse[1][1] + expansion
            alpha = ellipse[2]
            axis1 = min(axis1, max_extended_radius)
            axis2 = min(axis2, max_extended_radius)
            image = cv.ellipse(
                image,
                (round(ceny), round(cenx)),
                (round(axis1 / 2), round(axis2 / 2)),
                alpha,
                0,
                360,
                (0, 0, 22),  # in the RGB cube, set blue plane pixels of the ellipse to 22
                -1,
            )
            #  Create another non-extended ellipse that is used to create the
            #  persist_jumps for this integration. This will be used to mask groups
            #  in subsequent integrations.
            sat_ellipse = image[:, :, 2]  # extract the Blue plane of the image
            saty, satx = np.where(sat_ellipse == 22)  # find all the ellipse pixels in the ellipse
            outcube[grp:, saty, satx] = sat_flag
            persist_image = cv.ellipse(
                persist_image,
                (round(ceny), round(cenx)),
                (round(ellipse[1][0] / 2), round(ellipse[1][1] / 2)),
                alpha,
                0,
                360,
                (0, 0, 22),
                -1,
            )
            persist_ellipse = persist_image[:, :, 2]
            persist_saty, persist_satx = np.where(persist_ellipse == 22)
            persist_jumps[persist_saty, persist_satx] = jump_flag
    return outcube, persist_jumps


def extend_ellipses(
    gdq_cube,
    intg,
    grp,
    ellipses,
    sat_flag,
    jump_flag,
    expansion=1.9,
    expand_by_ratio=True,
    num_grps_masked=1,
    max_extended_radius=200,
):
    # For a given DQ plane it will use the list of ellipses to create
    #  expanded ellipses of pixels with
    # the jump flag set.
    out_gdq_cube = gdq_cube.copy()
    plane = gdq_cube[intg, grp, :, :].copy()
    ncols = plane.shape[1]
    nrows = plane.shape[0]
    image = np.zeros(shape=(nrows, ncols, 3), dtype=np.uint8)
    num_ellipses = len(ellipses)
    for ellipse in ellipses:
        ceny = ellipse[0][0]
        cenx = ellipse[0][1]
        # Expand the ellipse by the expansion factor. The number of pixels
        # added to both axes is
        # the number of pixels added to the minor axis. This prevents very
        # large flagged ellipses
        # with high axis ratio ellipses. The major and minor axis are not
        # always the same index.
        # Therefore, we have to test to find which is actually the minor axis.
        if expand_by_ratio:
            if ellipse[1][1] < ellipse[1][0]:
                axis1 = ellipse[1][0] + (expansion - 1.0) * ellipse[1][1]
                axis2 = ellipse[1][1] * expansion
            else:
                axis1 = ellipse[1][0] * expansion
                axis2 = ellipse[1][1] + (expansion - 1.0) * ellipse[1][0]
        else:
            axis1 = ellipse[1][0] + expansion
            axis2 = ellipse[1][1] + expansion
        axis1 = min(axis1, max_extended_radius)
        axis2 = min(axis2, max_extended_radius)
        alpha = ellipse[2]
        image = cv.ellipse(
            image,
            (round(ceny), round(cenx)),
            (round(axis1 / 2), round(axis2 / 2)),
            alpha,
            0,
            360,
            (0, 0, jump_flag),
            -1,
        )
        jump_ellipse = image[:, :, 2]
        ngrps = gdq_cube.shape[1]
        last_grp = find_last_grp(grp, ngrps, num_grps_masked)
        #  This loop will flag the number of groups
        for flg_grp in range(grp, last_grp):
            sat_pix = np.bitwise_and(gdq_cube[intg, flg_grp, :, :], sat_flag)
            saty, satx = np.where(sat_pix == sat_flag)
            jump_ellipse[saty, satx] = 0
            out_gdq_cube[intg, flg_grp, :, :] = np.bitwise_or(gdq_cube[intg, flg_grp, :, :], jump_ellipse)
    diff_cube = out_gdq_cube - gdq_cube
    return out_gdq_cube, num_ellipses

def find_last_grp(grp, ngrps, num_grps_masked):
    """
    Parameters
    ----------
    grp : int
        The location of the shower

    ngrps : int
        The number of groups in the integration

    num_grps_masked : int
        The requested number of groups to be flagged after the shower

    Returns
    -------
    last_grp : int
        The index of the last group to flag for the shower

    """
    num_grps_masked += 1
    last_grp = min(grp + num_grps_masked, ngrps)
    return last_grp

def find_circles(dqplane, bitmask, min_area):
    # Using an input DQ plane this routine will find the groups of pixels with at least the minimum
    # area and return a list of the minimum enclosing circle parameters.
    pixels = np.bitwise_and(dqplane, bitmask)
    contours, hierarchy = cv.findContours(pixels, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)
    bigcontours = [con for con in contours if cv.contourArea(con) >= min_area]
    return [cv.minEnclosingCircle(con) for con in bigcontours]


def find_ellipses(dqplane, bitmask, min_area):
    # Using an input DQ plane this routine will find the groups of pixels with
    # at least the minimum
    # area and return a list of the minimum enclosing ellipse parameters.
    pixels = np.bitwise_and(dqplane, bitmask)
    contours, hierarchy = cv.findContours(pixels, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)
    bigcontours = [con for con in contours if cv.contourArea(con) > min_area]
    # minAreaRect is used because fitEllipse requires 5 points and it is
    # possible to have a contour
    # with just 4 points.
    return [cv.minAreaRect(con) for con in bigcontours]


def make_snowballs(
    gdq,
    integration,
    group,
    jump_ellipses,
    sat_ellipses,
    next_sat_ellipses,
    low_threshold,
    high_threshold,
    min_sat_radius,
    expansion,
    sat_flag,
    jump_flag,
    max_extended_radius,
    persist_jumps,
):
    # This routine will create a list of snowballs (ellipses) that have the
    # center
    # of the saturation circle within the enclosing jump rectangle.
    snowballs = []
    num_groups = gdq.shape[1]
    for jump in jump_ellipses:
        if near_edge(jump, low_threshold, high_threshold):
            # if the jump ellipse is near the edge, do not require saturation in the
            # center of the jump ellipse
            snowballs.append(jump)
        else:
            for sat in sat_ellipses:
                if ((point_inside_ellipse(sat[0], jump) and jump not in snowballs)):
                    snowballs.append(jump)
            if group < num_groups - 1:
                # Is there saturation inside the jump in the next group?
                for next_sat in next_sat_ellipses:
                    if ((point_inside_ellipse(next_sat[0], jump)) and jump not in snowballs):
                        snowballs.append(jump)
    # extend the saturated ellipses that are larger than the min_sat_radius
    gdq[integration, :, :, :], persist_jumps[integration, :, :] = extend_saturation(
        gdq[integration, :, :, :],
        group,
        sat_ellipses,
        sat_flag,
        jump_flag,
        min_sat_radius,
        persist_jumps[integration, :, :],
        expansion=expansion,
        max_extended_radius=max_extended_radius,
    )

    return gdq, snowballs, persist_jumps


def point_inside_ellipse(point, ellipse):
    delta_center = np.sqrt((point[0] - ellipse[0][0]) ** 2 + (point[1] - ellipse[0][1]) ** 2)
    major_axis = max(ellipse[1][0], ellipse[1][1])

    return delta_center < major_axis


def near_edge(jump, low_threshold, high_threshold):
    #  This routing tests whether the center of a jump is close to the edge of
    # the detector. Jumps that are within the threshold will not require a
    # saturated core since this may be off the detector
    return (
        jump[0][0] < low_threshold
        or jump[0][1] < low_threshold
        or jump[0][0] > high_threshold
        or jump[0][1] > high_threshold
    )


# MIRI cosmic ray showers code
def find_faint_extended(
    indata,
    ingdq,
    pdq,
    readnoise_2d,
    nframes,
    minimum_sigclip_groups,
    dqflags,
    snr_threshold=1.3,
    min_shower_area=40,
    inner=1,
    outer=2,
    donotuse_flag = 1,
    sat_flag=2,
    jump_flag=4,
    ellipse_expand=1.1,
    num_grps_masked=25,
    max_extended_radius=200,
    min_diffs_for_shower=10,
    max_shower_amplitude=6,
):
    """
    Parameters
    ----------
      indata : float, 4D array
          Science array.
      gdq : int, 2D array
          Group dq array.
      readnoise_2d : float, 2D array
          Readnoise for all pixels.
      nframes : int
          The number frames that are averaged in the group.
      snr_threshold : float
          The signal-to-noise ratio threshold for detection of extended
          emission.
      min_shower_area : int
          The minimum area for a group of pixels to be flagged as a shower.
      inner : int
          The inner radius of the ring_2D_kernal used for the convolution.
      outer : int
          The outer radius of the ring_2D_kernal used for the convolution.
      sat_flag : int
          The integer value of the saturation flag.
      jump_flag : int
          The integer value of the jump flag
      ellipse_expand : float
          The relative increase in the size of the fitted ellipse to be
          applied to the shower.
      num_grps_masked : int
          The number of groups after the detected shower to be flagged as jump.
      max_extended_radius : int
          The upper limit for the extension of saturation and jump
      minimum_sigclip_groups : int
          The minimum number of groups to use sigma clipping.
      max_shower_amplitude : float
          The maximum amplitude of shower artifacts to correct in DN/group


    Returns
    -------
    gdq : int, 4D array
      updated group dq array.
    number_ellipse : int
    Total number of showers detected.

    """
    log.info("Flagging Showers")
    refpix_flag = dqflags["REFERENCE_PIXEL"]
    gdq = ingdq.copy()
    data = indata.copy()
    nints = data.shape[0]
    ngrps = data.shape[1]
    num_grps_donotuse = 0

    for integ in range(nints):
        for grp in range(ngrps):
            if np.all(np.bitwise_and(gdq[integ, grp, :, :], donotuse_flag)):
                num_grps_donotuse += 1
    total_diffs = nints * (ngrps - 1) - num_grps_donotuse
    if total_diffs < min_diffs_for_shower:
        log.warning("Not enough differences for shower detections")
        return ingdq, 0
    read_noise_2 = readnoise_2d**2
    jump_dnu_flag = jump_flag + donotuse_flag
    sat_dnu_flag = sat_flag + donotuse_flag
    data[gdq == jump_dnu_flag] = np.nan
    data[gdq == sat_dnu_flag] = np.nan
    data[gdq == sat_flag] = np.nan
    data[gdq == jump_flag] = np.nan
    data[gdq == donotuse_flag] = np.nan
    refy, refx = np.where(pdq == refpix_flag)
    gdq[:, :, refy, refx] = donotuse_flag
    first_diffs = np.diff(data, axis=1)

    all_ellipses = []

    first_diffs_masked = np.ma.masked_array(first_diffs, mask=np.isnan(first_diffs))
    warnings.filterwarnings("ignore")
    if nints >= minimum_sigclip_groups:
        mean, median, stddev = stats.sigma_clipped_stats(first_diffs_masked, sigma=5, axis=0)
    else:
        median_diffs = np.nanmedian(first_diffs_masked, axis=(0, 1))
        sigma = np.sqrt(np.abs(median_diffs) + read_noise_2 / nframes)

    for intg in range(nints):
        # calculate sigma for each pixel
        if nints < minimum_sigclip_groups:
            # The difference from the median difference for each group
            if intg > 0:
                e_jump = first_diffs_masked[intg] - median_diffs[np.newaxis, :, :]

                # SNR ratio of each diff.
                ratio = np.abs(e_jump) / sigma[np.newaxis, :, :]
            else:
                median_diffs = np.nanmedian(first_diffs_masked[intg], axis=0)
                sigma = np.sqrt(np.abs(median_diffs) + read_noise_2 / nframes)
                # The difference from the median difference for each group
                e_jump = first_diffs_masked[intg] - median_diffs[np.newaxis, :, :]
                # SNR ratio of each diff.
                ratio = np.abs(e_jump) / sigma[np.newaxis, :, :]
                median_diffs = np.nanmedian(first_diffs_masked, axis=(0, 1))
                sigma = np.sqrt(np.abs(median_diffs) + read_noise_2 / nframes)
        #  The convolution kernel creation
        ring_2D_kernel = Ring2DKernel(inner, outer)
        first_good_group = find_first_good_group(gdq[intg, :, :, :], donotuse_flag)
        for grp in range(first_good_group + 1, ngrps):
            if nints >= minimum_sigclip_groups:
                median_diffs = median[grp - 1]
                sigma = stddev[grp - 1]
                # The difference from the median difference for each group
                e_jump = first_diffs_masked[intg] - median_diffs[np.newaxis, :, :]
                # SNR ratio of each diff.
                ratio = np.abs(e_jump) / sigma[np.newaxis, :, :]
            masked_ratio = ratio[grp - 1].copy()

            #  mask pixels that are already flagged as jump
            combined_pixel_mask = np.bitwise_or(gdq[intg, grp, :, :], pdq[:, :])
            jump_pixels_array = np.bitwise_and(combined_pixel_mask, jump_flag)
            jumpy, jumpx = np.where(jump_pixels_array == jump_flag)
            masked_ratio[jumpy, jumpx] = np.nan

            #  mask pixels that are already flagged as sat.
            sat_pixels_array = np.bitwise_and(combined_pixel_mask, sat_flag)
            saty, satx = np.where(sat_pixels_array == sat_flag)
            masked_ratio[saty, satx] = np.nan

            #  mask pixels that are already flagged as do not use
            dnu_pixels_array = np.bitwise_and(combined_pixel_mask, 1)
            dnuy, dnux = np.where(dnu_pixels_array == 1)
            masked_ratio[dnuy, dnux] = np.nan

            masked_smoothed_ratio = convolve(masked_ratio.filled(np.nan), ring_2D_kernel)
            #  mask out the pixels that got refilled by the convolution
            masked_smoothed_ratio[dnuy, dnux] = np.nan
            masked_smoothed_ratio[saty, satx] = np.nan
            masked_smoothed_ratio[jumpy, jumpx] = np.nan
            nrows = ratio.shape[1]
            ncols = ratio.shape[2]
            extended_emission = np.zeros(shape=(nrows, ncols), dtype=np.uint8)
            exty, extx = np.where(masked_smoothed_ratio > snr_threshold)
            extended_emission[exty, extx] = 1

            #  find the contours of the extended emission
            contours, hierarchy = cv.findContours(extended_emission, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)
            #  get the contours that are above the minimum size
            bigcontours = [con for con in contours if cv.contourArea(con) > min_shower_area]
            #  get the minimum enclosing rectangle which is the same as the
            # minimum enclosing ellipse
            ellipses = [cv.minAreaRect(con) for con in bigcontours]
            expand_by_ratio = True
            expansion = 1.0
            plane = gdq[intg, grp, :, :]
            nrows = plane.shape[0]
            ncols = plane.shape[1]
            image = np.zeros(shape=(nrows, ncols, 3), dtype=np.uint8)
            image2 = np.zeros_like(image)
            cv.drawContours(image2, bigcontours, -1, (0, 0, jump_flag), -1)
            for ellipse in ellipses:
                ceny = ellipse[0][0]
                cenx = ellipse[0][1]
                # Expand the ellipse by the expansion factor. The number of pixels
                # added to both axes is
                # the number of pixels added to the minor axis. This prevents very
                # large flagged ellipses
                # with high axis ratio ellipses. The major and minor axis are not
                # always the same index.
                # Therefore, we have to test to find which is actually the minor axis.
                if expand_by_ratio:
                    if ellipse[1][1] < ellipse[1][0]:
                        axis1 = ellipse[1][0] + (expansion - 1.0) * ellipse[1][1]
                        axis2 = ellipse[1][1] * expansion
                    else:
                        axis1 = ellipse[1][0] * expansion
                        axis2 = ellipse[1][1] + (expansion - 1.0) * ellipse[1][0]
                else:
                    axis1 = ellipse[1][0] + expansion
                    axis2 = ellipse[1][1] + expansion
                axis1 = min(axis1, max_extended_radius)
                axis2 = min(axis2, max_extended_radius)
                alpha = ellipse[2]
                image = cv.ellipse(
                    image,
                    (round(ceny), round(cenx)),
                    (round(axis1 / 2), round(axis2 / 2)),
                    alpha,
                    0,
                    360,
                    (0, 0, jump_flag),
                    -1,
                )
            if len(ellipses) > 0:
                # add all the showers for this integration to the list
                all_ellipses.append([intg, grp, ellipses])
                # Reset the warnings filter to its original state
    warnings.resetwarnings()
    total_showers = 0

    if all_ellipses:
        #  Now we actually do the flagging of the pixels inside showers.
        # This is deferred until all showers are detected. because the showers
        # can flag future groups and would confuse the detection algorithm if
        # we worked on groups that already had some flagged showers.
        for showers in all_ellipses:
            intg = showers[0]
            grp = showers[1]
            ellipses = showers[2]
            total_showers += len(ellipses)
            gdq, num = extend_ellipses(
                gdq,
                intg,
                grp,
                ellipses,
                sat_flag,
                jump_flag,
                expansion=ellipse_expand,
                expand_by_ratio=True,
                num_grps_masked=num_grps_masked,
                max_extended_radius=max_extended_radius
            )

    # Ensure that flagging showers didn't change final fluxes by more than the allowed amount
    for intg in range(nints):
        # Consider DO_NOT_USE, SATURATION, and JUMP_DET flags
        invalid_flags = donotuse_flag | sat_flag | jump_flag

        # Approximate pre-shower rates
        tempdata = indata[intg, :, :, :].copy()
        # Ignore any groups flagged in the original gdq array
        tempdata[ingdq[intg, :, :, :] & invalid_flags != 0] = np.nan
        # Compute group differences
        diff = np.diff(tempdata, axis=0)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning, message="All-NaN")
            warnings.filterwarnings("ignore", category=RuntimeWarning, message="Mean of empty slice")
            image1 = np.nanmean(diff, axis=0)

        # Approximate post-shower rates
        tempdata = indata[intg, :, :, :].copy()
        # Ignore any groups flagged in the shower gdq array
        tempdata[gdq[intg, :, :, :] & invalid_flags != 0] = np.nan
        # Compute group differences
        diff = np.diff(tempdata, axis=0)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning, message="All-NaN")
            warnings.filterwarnings("ignore", category=RuntimeWarning, message="Mean of empty slice")
            image2 = np.nanmean(diff, axis=0)

        # Revert the group flags to the pre-shower flags for any pixels whose rates
        # became NaN or changed by more than the amount reasonable for a real CR shower
        # Note that max_shower_amplitude should now be in DN/group not DN/s
        diff = np.abs(image1 - image2)
        indx = np.where((np.isfinite(diff) == False) | (diff > max_shower_amplitude))
        gdq[intg, :, indx[0], indx[1]] = ingdq[intg, :, indx[0], indx[1]]

    return gdq, total_showers

def find_first_good_group(int_gdq, do_not_use):
    ngrps = int_gdq.shape[0]
    skip_grp = True
    first_good_group = 0
    for grp in range(ngrps):
        mask = np.bitwise_and(int_gdq[grp], do_not_use)
        skip_grp = np.all(mask)
        if not skip_grp:
            first_good_group = grp
            break
    return first_good_group

def calc_num_slices(n_rows, max_cores, max_available):
    n_slices = 1
    if max_cores.isnumeric():
        n_slices = int(max_cores)
    elif max_cores.lower() == "none" or max_cores.lower() == "one":
        n_slices = 1
    elif max_cores == "quarter":
        n_slices = max_available // 4 or 1
    elif max_cores == "half":
        n_slices = max_available // 2 or 1
    elif max_cores == "all":
        n_slices = max_available
    # Make sure we don't have more slices than rows or available cores.
    return min([n_rows, n_slices, max_available])



from astropy.modeling.rotations import RotationSequence3D

__all__ = ["_wcsinfo_from_wcs_transform"]


def _wcsinfo_from_wcs_transform(wcs):
    frames = wcs.available_frames
    if "v2v3" not in frames or "world" not in frames or frames[-1] != "world":
        msg = "Unsupported WCS structure."
        raise ValueError(msg)

    # Initially get v2_ref, v3_ref, and roll_ref from
    # the v2v3 to world transform. Also get ra_ref, dec_ref
    t = wcs.get_transform(frames[-2], "world")
    for m in t:
        if isinstance(m, RotationSequence3D) and m.parameters.size == 5:
            v2_ref, nv3_ref, roll_ref, dec_ref, nra_ref = m.angles.value
            break
    else:
        msg = "Unsupported WCS structure."
        raise ValueError(msg)

    # overwrite v2_ref, v3_ref, and roll_ref with
    # values from the tangent plane when available:
    if "v2v3corr" in frames:
        # get v2_ref, v3_ref, and roll_ref from
        # the v2v3 to v2v3corr transform:
        frm1 = "v2v3vacorr" if "v2v3vacorr" in frames else "v2v3"
        tpcorr = wcs.get_transform(frm1, "v2v3corr")
        v2_ref, nv3_ref, roll_ref = tpcorr["det_to_optic_axis"].angles.value

    return {
        "v2_ref": 3600 * v2_ref,
        "v3_ref": -3600 * nv3_ref,
        "roll_ref": roll_ref,
        "ra_ref": -nra_ref,
        "dec_ref": dec_ref
    }



import warnings

warnings.warn(
    "dynamicdq has been moved to stdatamodels.dynamicdq, please use that instead",
    DeprecationWarning,
    stacklevel=2,
)

try:
    from stdatamodels.dynamicdq import dynamic_mask
except ImportError as err:
    msg = "dynamicdq has been moved to stdatamodels.dynamicdq, please install stdatamodels"
    raise ImportError(msg) from err


__all__ = ["dynamic_mask"]



"""
Unit tests for dark current correction
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose

from stcal.dark_current.dark_class import DarkData, ScienceData
from stcal.dark_current.dark_sub import average_dark_frames_3d as average_dark_frames
from stcal.dark_current.dark_sub import do_correction_data as darkcorr

dqflags = {
    "DO_NOT_USE": 2**0,  # Bad pixel. Do not use.
    "SATURATED": 2**1,  # Pixel saturated during exposure
    "JUMP_DET": 2**2,  # Jump detected during exposure
    "DROPOUT": 2**3,  # Data lost in transmission
    "AD_FLOOR": 2**6,  # Below A/D floor (0 DN, was RESERVED_3)
}


# Define frame_time and number of groups in the generated dark reffile
TFRAME = 10.73677
NGROUPS_DARK = 10

DELIM = "-" * 80


@pytest.fixture()
def make_rampmodel():
    """Make MIRI Ramp model for testing"""

    def _ramp(nints, ngroups, nrows, ncols):
        ramp_data = ScienceData()

        dims = (nints, ngroups, nrows, ncols)
        imshape = (nrows, ncols)

        ramp_data.data = np.full(dims, 1.0, dtype=np.float32)
        ramp_data.groupdq = np.zeros(dims, dtype=np.uint32)
        ramp_data.pixeldq = np.zeros(imshape, dtype=np.uint32)
        ramp_data.err = np.zeros(dims, dtype=np.float32)

        ramp_data.instrument_name = "MIRI"

        ramp_data.cal_step = None

        return ramp_data

    return _ramp


@pytest.fixture()
def make_darkmodel():
    """Make MIRI dark model for testing"""

    def _dark(ngroups, nrows, ncols):
        # create the data and groupdq arrays
        nints = 2
        csize = (nints, ngroups, nrows, ncols)

        dark = DarkData()
        dark.data = np.full(csize, 1.0)
        dark.groupdq = np.zeros(csize, dtype=np.uint32)
        dark.err = np.zeros(csize, dtype=np.float32)

        dark.exp_nframes = 1
        dark.exp_ngroups = ngroups
        dark.exp_groupgap = 0

        return dark

    return _dark


@pytest.fixture()
def setup_nrc_cube():
    """Set up fake NIRCam data to test."""

    def _cube(readpatt, ngroups, nframes, groupgap, nrows, ncols):
        nints = 1
        dims = (nints, ngroups, nrows, ncols)
        ramp_data = ScienceData()
        ramp_data.data = np.zeros(dims, dtype=np.float32)
        ramp_data.groupdq = np.zeros(dims, dtype=np.uint32)
        ramp_data.pixeldq = np.zeros(dims, dtype=np.uint32)
        ramp_data.err = np.zeros(dims, dtype=np.float32)

        ramp_data.instrument_name = "NIRCAM"
        ramp_data.exp_nframes = nframes
        ramp_data.exp_groupgap = groupgap

        ramp_data.cal_step = None

        dark_dims = (NGROUPS_DARK, 2048, 2048)
        dark_data = DarkData(dark_dims)
        dark_data.exp_nframes = nframes
        dark_data.exp_ngroups = NGROUPS_DARK
        dark_data.exp_groupgap = 0

        return ramp_data, dark_data

    return _cube


def _params():
    """Returns list of tuples, one for each readpatt, generating parameters for
    test_frame_averaging. Parameters are the following:

        (readpatt, ngroups, nframes, groupgap, nrows, ncols)

    Note groupgap = nskip
    """

    # Dictionary of NIRCam readout patterns
    readpatterns = {
        "DEEP8": {"ngroups": 20, "nframes": 8, "nskip": 12},
        "DEEP2": {"ngroups": 20, "nframes": 2, "nskip": 18},
        "MEDIUM8": {"ngroups": 10, "nframes": 8, "nskip": 2},
        "MEDIUM2": {"ngroups": 10, "nframes": 2, "nskip": 8},
        "SHALLOW4": {"ngroups": 10, "nframes": 4, "nskip": 1},
        "SHALLOW2": {"ngroups": 10, "nframes": 2, "nskip": 3},
        "BRIGHT2": {"ngroups": 10, "nframes": 2, "nskip": 0},
        "BRIGHT1": {"ngroups": 10, "nframes": 1, "nskip": 1},
        "RAPID": {"ngroups": 10, "nframes": 1, "nskip": 0},
    }

    params = []
    ngroups = 3
    # NIRCam is 2048x2048, but we reduce the ncols to 20x20 for speed/memory
    nrows = 20
    ncols = 20
    for readpatt, values in readpatterns.items():
        params.append((readpatt, ngroups, values["nframes"], values["nskip"], nrows, ncols))

    return params


@pytest.mark.parametrize(("readpatt", "ngroups", "nframes", "groupgap", "nrows", "ncols"), _params())
def test_frame_averaging(setup_nrc_cube, readpatt, ngroups, nframes, groupgap, nrows, ncols):
    """Check that if nframes>1 or groupgap>0, then the pipeline reconstructs
    the dark reference file to match the frame averaging and groupgap
    settings of the exposure."""

    # Create data and dark model
    data, dark = setup_nrc_cube(readpatt, ngroups, nframes, groupgap, nrows, ncols)

    # Add ramp values to dark model data array
    dark.data[:, 10, 10] = np.arange(0, NGROUPS_DARK)
    dark.err[:, 10, 10] = np.arange(10, NGROUPS_DARK + 10)

    # Run the pipeline's averaging function
    avg_dark = average_dark_frames(dark, ngroups, nframes, groupgap)

    # Group input groups into collections of frames which will be averaged
    total_frames = (nframes * ngroups) + (groupgap * (ngroups - 1))

    # Get starting/ending indexes of the input groups to be averaged
    gstrt_ind = np.arange(0, total_frames, nframes + groupgap)
    gend_ind = gstrt_ind + nframes

    # Prepare arrays to hold results of averaging
    manual_avg = np.zeros((ngroups), dtype=np.float32)
    manual_errs = np.zeros((ngroups), dtype=np.float32)

    # Manually average the input data to compare with pipeline output
    for newgp, gstart, gend in zip(range(ngroups), gstrt_ind, gend_ind):
        # Average the data frames
        newframe = np.mean(dark.data[gstart:gend, 10, 10])
        manual_avg[newgp] = newframe

        # ERR arrays will be quadratic sum of error values
        manual_errs[newgp] = np.sqrt(np.sum(dark.err[gstart:gend, 10, 10] ** 2)) / (gend - gstart)

    # Check that pipeline output matches manual averaging results
    assert_allclose(manual_avg, avg_dark.data[:, 10, 10], rtol=1e-5)
    assert_allclose(manual_errs, avg_dark.err[:, 10, 10], rtol=1e-5)

    # Check that meta data was properly updated
    assert avg_dark.exp_nframes == nframes
    assert avg_dark.exp_ngroups == ngroups
    assert avg_dark.exp_groupgap == groupgap


def test_more_sci_frames(make_rampmodel, make_darkmodel):
    """
    Check that data is unchanged if there are more frames in the science
    data is than in the dark reference file and verify that when the dark
    is not applied, the data is correctly flagged as such
    """

    # ncols of integration
    nints, ngroups, nrows, ncols = 1, 7, 200, 200

    # create raw input data for step
    dm_ramp = make_rampmodel(nints, ngroups, nrows, ncols)
    dm_ramp.exp_nframes = 1
    dm_ramp.exp_groupgap = 0

    # populate data array of science cube
    for i in range(ngroups - 1):
        dm_ramp.data[0, i] = i

    refgroups = 5
    # create dark reference file model with fewer frames than science data
    dark = make_darkmodel(refgroups, nrows, ncols)

    # populate data array of reference file
    for i in range(refgroups - 1):
        dark.data[0, i] = i * 0.1

    # apply correction
    out_data, avg_data = darkcorr(dm_ramp, dark)

    # test that the science data are not changed; input file = output file
    np.testing.assert_array_equal(dm_ramp.data, out_data.data)

    # get dark correction status from header
    # darkstatus = outfile.meta.cal_step.dark_sub
    darkstatus = out_data.cal_step

    assert darkstatus == "SKIPPED"


def test_sub_by_frame(make_rampmodel, make_darkmodel):
    """
    Check that if NFRAMES=1 and GROUPGAP=0 for the science data,
    the dark reference data are directly subtracted frame by frame
    """

    # size of integration
    nints, ngroups, nrows, ncols = 1, 10, 200, 200

    # create raw input data for step
    dm_ramp = make_rampmodel(nints, ngroups, nrows, ncols)
    dm_ramp.exp_nframes = 1
    dm_ramp.exp_groupgap = 0

    # populate data array of science cube
    for i in range(ngroups - 1):
        dm_ramp.data[0, i] = i

    # create dark reference file model with more frames than science data
    refgroups = 15
    dark = make_darkmodel(refgroups, nrows, ncols)

    # populate data array of reference file
    for i in range(refgroups - 1):
        dark.data[0, i] = i * 0.1

    # apply correction
    outfile, avg_dark = darkcorr(dm_ramp, dark)

    assert outfile.cal_step == "COMPLETE"

    # remove the single dimension at start of file (1, 30, 1032, 1024)
    # so comparison in assert works
    outdata = np.squeeze(outfile.data)

    # check that the dark file is subtracted frame by frame
    # from the science data
    diff = dm_ramp.data[0] - dark.data[0, :ngroups]

    # test that the output data file is equal to the difference
    # found when subtracting ref file from sci file
    tol = 1.0e-6
    np.testing.assert_allclose(outdata, diff, tol)


def test_nan(make_rampmodel, make_darkmodel):
    """
    Verify that when a dark has NaNs, these are correctly
    assumed as zero and the PIXELDQ is set properly
    """

    # size of integration
    nints, ngroups, nrows, ncols = 1, 10, 200, 200

    # create raw input data for step
    dm_ramp = make_rampmodel(nints, ngroups, nrows, ncols)
    dm_ramp.exp_nframes = 1
    dm_ramp.exp_groupgap = 0

    # populate data array of science cube
    for i in range(ngroups - 1):
        dm_ramp.data[0, i, :, :] = i

    # create dark reference file model with more frames than science data
    refgroups = 15
    dark = make_darkmodel(refgroups, nrows, ncols)

    # populate data array of reference file
    for i in range(refgroups - 1):
        dark.data[0, i] = i * 0.1

    # set NaN in dark file

    dark.data[0, 5, 100, 100] = np.nan

    # apply correction
    outfile, avg_dark = darkcorr(dm_ramp, dark)

    # test that the NaN dark reference pixel was set to 0 (nothing subtracted)
    assert outfile.data[0, 5, 100, 100] == 5.0


def test_dq_combine(make_rampmodel, make_darkmodel):
    """
    Verify that the DQ array of the dark is correctly
    combined with the PIXELDQ array of the science data.
    """

    # size of integration
    nints, ngroups, nrows, ncols = 1, 5, 200, 200

    # create raw input data for step
    dm_ramp = make_rampmodel(nints, ngroups, nrows, ncols)
    dm_ramp.exp_nframes = 1
    dm_ramp.exp_groupgap = 0

    # populate data array of science cube
    for i in range(1, ngroups - 1):
        dm_ramp.data[0, i, :, :] = i

    # create dark reference file model with more frames than science data
    refgroups = 7
    dark = make_darkmodel(refgroups, nrows, ncols)

    # populate dq flags of sci pixeldq and reference dq
    dm_ramp.pixeldq[50, 50] = dqflags["JUMP_DET"]
    dm_ramp.pixeldq[50, 51] = dqflags["SATURATED"]

    dark.groupdq[0, 0, 50, 50] = dqflags["DO_NOT_USE"]
    dark.groupdq[0, 0, 50, 51] = dqflags["DO_NOT_USE"]

    # run correction step
    outfile, avg_dark = darkcorr(dm_ramp, dark)

    # check that dq flags were correctly added
    assert outfile.pixeldq[50, 50] == np.bitwise_or(dqflags["JUMP_DET"], dqflags["DO_NOT_USE"])

    assert outfile.pixeldq[50, 51] == np.bitwise_or(dqflags["SATURATED"], dqflags["DO_NOT_USE"])


def test_frame_avg(make_rampmodel, make_darkmodel):
    """
    Check that if NFRAMES>1 or GROUPGAP>0, the frame-averaged dark data are
    subtracted group-by-group from science data groups and the ERR arrays
    are not modified
    """

    # size of integration
    nints, ngroups, nrows, ncols = 1, 5, 1024, 1032

    # create raw input data for step
    dm_ramp = make_rampmodel(nints, ngroups, nrows, ncols)
    dm_ramp.exp_nframes = 4
    dm_ramp.exp_groupgap = 0

    # populate data array of science cube
    for i in range(ngroups - 1):
        dm_ramp.data[:, i] = i + 1

    # create dark reference file model

    refgroups = 20  # This needs to be 20 groups for the calculations to work
    dark = make_darkmodel(refgroups, nrows, ncols)

    # populate data array of reference file
    for i in range(refgroups - 1):
        dark.data[0, i] = i * 0.1

    # apply correction
    outfile, avg_dark = darkcorr(dm_ramp, dark)

    # dark frames should be averaged in groups of 4 frames

    # this will result in average values of 0.15, 0.55, 0.95, and 1.35
    # these values are then subtracted from frame values of 1, 2, 3 and 4

    assert outfile.data[0, 0, 500, 500] == pytest.approx(0.85)
    assert outfile.data[0, 1, 500, 500] == pytest.approx(1.45)
    assert outfile.data[0, 2, 500, 500] == pytest.approx(2.05)
    assert outfile.data[0, 3, 500, 500] == pytest.approx(2.65)

    # check that the error array is not modified.
    tol = 1.0e-6
    np.testing.assert_allclose(outfile.err[:, :], 0, tol)
